#!/bin/python
# rubika client user self methods

import warnings, sys, os, inspect, typing
from requests       import get, post, session
from pathlib        import Path
from random         import choice, randint, sample
from re             import findall, search, sub, compile, escape
from time           import sleep, gmtime, localtime
from datetime       import datetime
from json           import load, loads, dumps
from .crypto        import Encryption
from .connection    import GetData, Urls
from .clients       import *
from .exceptions    import *
from .extensions    import *
from .sessions      import RubikaClient as Client
from .parser        import (MessageEmpty, MessageEntityBold,
                            MessageEntityCode, MessageEntityItalic,
                            MessageEntityHashtag, MessageEntityMention,
                            MessageEntityMentionName, MessageEntityPre,
                            MessageEntityStrike, MessageEntityTextUrl,
                            MessageEntityUnderline, MessageEntityUnknown,
                            MessageEntityUrl, Metas, Tags, MetaDataLoader)


class UserMethods:

    __extend_class__ = 'UserMethods'

    # to use the `with`
    def __enter__(self: Client):
        return (self)

    def __exit__(
        self: Client, *args,
        **kwargs) -> (None):
        pass

    # new shortcut to use the first important methods
    # to usage of operators

    def __eq__(self, kwargs):

        # Example:
        '''
        from rb import RubikaClient
        
        with RubikaClient(...) as client:
            print(client == dict(chat_id=..., text=...)) # to send_message
        
        shortcut:
            self == dict(chat_id=..., text=...)
        '''

        if not isinstance(kwargs, dict):
            kwargs: dict = kwargs.dict()
        
        return self.send_message(**kwargs)

    def __ge__(self, kwargs):
        '''
        self >= dict(chat_id=..., text=...)
        '''

        if not isinstance(kwargs, dict):
            kwargs: dict = kwargs.dict()

        return self.edit_message(**kwargs)

    def __gt__(self, *args, **kwargs):
        '''
        self > 'chat-guid'
        '''

        return self.get_chat_last_message(*args, **kwargs)

    def __ne__(self, kwargs):
        '''
        self != dict(message_ids=[...,], chat_id=...)
        '''

        if not isinstance(kwargs, dict):
            kwargs: dict = kwargs.dict()
            
        return self.delete_messages(**kwargs)

    def __lt__(self, kwargs):
        '''
        self < dict(message_id=..., chat_id=...)
        '''

        return self.set_pin_message(**kwargs)

    def __le__(self, *args):
        '''
        self <= {'chat':'message'}
        '''

        return self.seen_chats(*args)

    def __add__(self, kwargs):
        '''
        self + dict(from_object=..., message_ids=[], to=...)
        '''

        if not isinstance(kwargs, dict):
            kwargs: dict = kwargs.dict()
            
        return self.forward_messages(**kwargs)

    def __sub__(self, kwargs):
        '''
        self - dict(message={})
        '''

        return self.get_file(**kwargs)

    def __mul__(self, *args):
        '''
        self * 'chat-guid'
        '''

        return self.get_chat_info(*args)

    def __floordiv__(self, kwargs):
        '''
        self // dict(...)
        '''

        return self.send_movie(**kwargs)

    def __truediv__(self, kwargs):
        '''
        self / dict(...)
        '''

        return self.send_music(**kwargs)

    def __mod__(self, kwargs):
        '''
        self % dict(...)
        '''

        return self.send_voice(**kwargs)

    def __pow__(self, kwargs):
        '''
        self ** dict(...)
        '''

        return self.send_document(**kwargs)

    def __lshift__(self, kwargs):
        '''
        self << dict(...)
        '''

        return self.send_location(**kwargs)

    def __rshift__(self, kwargs):
        '''
        self >> dict(chat_id=..., action='Block')
        '''
        return self.set_block_user(**kwargs)

    def __and__(self, *args):
        '''
        self & dict(...)
        '''
        
        return self.get_chat_last_message_id(*args)

    def __or__(self, *args):
        '''
        self | dict(...)
        '''
        
        return self.get_chat_last_text(*args)

    def __xor__(self, *args):
        '''
        self ^ 'chat-guid'
        '''
        
        return self.get_chat_last_object(*args)

    
    # lambda func methods.

    # this method for logout and delete sessions
    log_out                     =   lambda self='Client': GetData.api(version='5', auth=self.auth, method='logout', mode=self.city, proxy=self.proxy, data={}, platform='web')

    # this methpd for getting all chats
    get_chats                   =   lambda self='Client', start_id=None: GetData.api(version=self.api_version or '5', auth=self.auth, method='getChats', data={'start_id': start_id}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)

    # this method for geting new chats
    get_chats_updates           =   lambda self='Client', state=None: GetData.api(version=self.api_version or '5', auth=self.auth, method='getChatsUpdates', data={'state'  :   state or str(round(datetime.today().timestamp()) - 200)}, mode=self.city, proxy=self.proxy, platform=self.platform or 'web')

    # it method is for get stickers 
    my_sticker_set              =   lambda self='Client': GetData.api(version=self.api_version or '5', method='getMyStickerSets', auth=self.auth, data={}, mode=self.city, proxy=self.proxy, platform=self.platform or 'web')

    # it method is for get account all folders
    get_folders                 =   lambda self='Client': GetData.api(version=self.api_version or '5', auth=self.auth, method='getFolders', data={}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)

    # setting methods
    get_privacy_setting         =   lambda self=Client: GetData.api(version=self.api_version or '5', method='getPrivacySetting', auth=self.auth, data={}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    get_my_sessions             =   lambda self=Client: GetData.api(version=self.api_version or '5', method='getMySessions', auth=self.auth, data={}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    get_two_passcode_status     =   lambda self=Client: GetData.api(version=self.api_version or '5', method='getTwoPasscodeStatus', auth=self.auth, data={}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    check_two_step_passcode     =   lambda self, password: GetData.api(version=self.api_version or '5', method='checkTwoStepPasscode', auth=self.auth, data={'password': password}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    change_password             =   lambda self, password, new_password, new_hint: GetData.api(version=self.api_version or '5', method='changePassword', auth=self.auth, data={'password': password, 'new_password': new_password, 'new_hint': new_hint}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    turn_off_two_step           =   lambda self, password: GetData.api(version=self.api_version or '5', method='turnOffTwoStep', auth=self.auth, data={'password': password}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    setup_two_step_verification =   lambda self, password, hint: GetData.api(version=self.api_version or '5', method='setupTwoStepVerification', auth=self.auth, data={'password': password, 'hint': hint}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    get_block_users             =   lambda self: GetData.api(version=self.api_version or '5', method='getBlockUsers', auth=self.auth, data={}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    request_recovery_email      =   lambda self, password, recovery_email: GetData.api(version=self.api_version or '5', method='requestRecoveryEmail', auth=self.auth, data={'password': password, 'recovery_email': recovery_email}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    verify_recovery_email       =   lambda self, password, code: GetData.api(version=self.api_version or '5', method='verifyRecoveryEmail', auth=self.auth, data={'password': password, 'code': code}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    get_contacts_updates        =   lambda self='Client', state=str(round(datetime.today().timestamp()) - 2000): GetData.api(version=self.api_version or '5', method='getContactsUpdates', auth=self.auth, data={'state': state}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    get_contacts                =   lambda self=Client : GetData.api(version=self.api_version or '5', method='getContacts', auth=self.auth, data={}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    get_suggested_folders       =   lambda self=Client : GetData.api(version=self.api_version or '5', method='getSuggestedFolders', auth=self.auth, data={}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    get_chat_last_author        =   lambda self, chat_id: self.get_chat_info(chat_id).get('data').get('chat').get('last_message').get('author_object_guid')


    @classmethod
    @property
    def key_generation(cls) -> (str):

        '''
        this method for create a fake session key
        
        `print(self.key_generation)`
        '''

        key: (str) = ''.join(__import__('random').sample(str('qweertyuiopasdfghjklzxcvbnm' * 2), 32))

        return ''.join(list(map(lambda a: chr((a - ord('a') + ord('\t')) % ord('\x1a') + ord('a')), (str(key[16:24] + key[:8] + key[24:32] + key[8:16]).encode('latin-1')))))


    @classmethod
    def tmp_generation(cls) -> (str):
        
        '''
        this method for creating a tmp session
        '''

        return ''.join(list(map(lambda i: choice([*__import__('string').ascii_lowercase, *str(__import__('string').digits)]), range(32))))


    @classmethod
    def private_link_generation(cls,
                  __type: typing.Literal['channel', 'group'] = 'group',
                  /, *args, **kwargs) -> str:

        '''
        this is has a method for create a link with hash for rubika
        
        `print(self.private_link_generation('group'))`
        '''

        hash = sample(__import__('string').ascii_uppercase * 2, 31)
        hash.insert(randint(16, 25), '0')

        return 'https://rubika.ir/join{}/'.format(next(iter([*__type.lower()]))) + ''.join(hash)


    @classmethod
    def guid_generation(
        cls,  __type: typing.Literal['channel', 'service',
                                     'group', 'user', 'bot'] = 'group',
        /, *args, **kwargs,) -> str:

        '''
        ### what is guid on rubika?
            - global unique identifier, has for chat id | created by md5 hashes & other ...
        
        ## this is a method for create a fake rubika guid
        
        `self.guid_generation('group') # g0B871703adf864fa497933c19afaede `
        '''

        string: str = iter([*__type.lower()]).__next__() + '0' + __import__('uuid').uuid4().__str__().replace('-', '')

        return string[: 1] + string[1: 5].upper() + string[5: string.__len__() - 2]


    @classmethod
    def send_code(
        cls,
        phone_number    :   (str),
        send_type       :   (str)   =   ('SMS'),
        password        :   (str)   =   None
        ) -> (dict):

        '''
        send_type <- key/value -> SMS/Internal
        '''

        data: dict = {'phone_number': f'98{phone_number[1:]}', 'send_type': send_type}

        if password:
            data.update({'pass_key': password})
        
        return (GetData.tmp(tmp=cls.key_generation, method='sendCode', data=data, mode='mashhad', proxy={'http': 'http://127.0.0.1:9050'}))


    @classmethod
    def sign_up(cls, *args, **kwargs) -> typing.Union[dict, str]:

        '''
        `self.sign_up('09...')`
        '''

        return cls.send_code(*args, **kwargs)


    @classmethod
    def sign_in(
        cls,
        phone_number    :   (str),
        phone_code_hash :   (str),
        phone_code      :   (str)
        ) -> typing.Union[dict, str]:

        '''
        this method for login or signin with lib

        PARAMETERS:
            1- phone_number : phone number of target's account : 09XXXXXXXXX
            2- phone_code_hash : hash of code sent to phone
            3- phone_code : code sent to phone
        '''

        return GetData.tmp(
            method='signIn',
            mode='mashhad',
            tmp=(cls.key_generation),
            proxy={'http': 'http://127.0.0.1:9050'},
            data={
                'phone_number': f'98{phone_number[1:]}',
                'phone_code_hash': phone_code_hash,
                'phone_code': phone_code
                }
            )


    @classmethod
    def log_in(cls, *args, **kwargs) -> typing.Union[dict, str]:
        '''
        `self.log_in('09...', 'hash', '12345')`
        '''

        return cls.sign_in(*args, **kwargs)


    @staticmethod
    def register_device(auth: (str),
                        device: (dict) = Device.DEFAULT_DEVICE) -> typing.Union[dict, str]:
        
        '''
        this method for registering your acocunt
        '''

        return GetData.api(
            version='4',
            auth=auth,
            method='registerDevice',
            mode='mashhad',
            platform='android',
            proxy={'http': 'http://127.0.0.1:9050'},
            data=(device)
            )


    @staticmethod
    def parsation (
        mode    : (str),
        text    : (str),
        user_ids: list = None
        ) -> typing.Tuple[str, list]:

        '''
        mode: to use a action type: html, markdown

        markdown [text]:

            **: Bold Type
            __: Italic Type
            ~~: Strike Type
            ``: Mono Type
            ```: Mono Type For Code
            @: To Mention Text

            Excample: '**Hey** ! @My Friend@ __i love__ ``you`` \n ```codes```'
        
        user_ids: to use the mention: '@You@ and @you@', ['u0...', 'u0...']
        '''

        results: list = []

        if (mode.upper().startswith('HTML')):

            realText: (str)  = Clean.html_cleaner(text)
            bolds   : (list) = findall(r'<b>(.*?)</b>', text)
            italics : (list) = findall(r'<i>(.*?)</i>', text)
            monos   : (list) = findall(r'<pre>(.*?)</pre>', text)
            bResult : (list) = [realText.index(i) for i in bolds]
            iResult : (list) = [realText.index(i) for i in italics]
            mResult : (list) = [realText.index(i) for i in monos]

            [results.append({'from_index': bIndex, 'length': len(bWord), 'type': 'Bold'}) for bIndex, bWord in zip(bResult, bolds)]
            [results.append({'from_index': iIndex, 'length': len(iWord), 'type': 'Italic'}) for iIndex, iWord in zip(iResult, italics)]
            [results.append({'from_index': mIndex, 'length': len(mWord), 'type': 'Mono'}) for mIndex, mWord in zip(mResult, monos)]

            return results, realText

        elif (mode.lower().startswith('markdown')):
            return Metas(text, user_ids=user_ids).checker


    @property
    def check_auth(self) -> (dict):

        '''
        this method for checking your auth ke
        '''

        try:
            GetData.api(version='4', auth=self.auth, method='', data={}, mode=self.city, proxy=self.proxy, platform='android')
            return {'status': 'ok'}
        
        except Exception:
            return {'status': 'error'}


    def request_send_file(self, file: (str)) -> typing.Union[dict, str]:
        
        return GetData.api(
            version='5',
            method='requestSendFile',
            auth=self.auth,
            mode=self.city,
            platform='web',
            data={
                'file_name': str(file.split('/')[-1]),
                'mime': file.split('.')[-1],
                'size': Path(file).stat().st_size
                }
            )


    def upload_file(self, file: str,
                    mode: bool = False) -> typing.Tuple[dict, dict]:

        frequest: dict = self.request_send_file(file).get('data')
            
        if (not 'http' in file):
            
            btytef = open(file, 'rb').read()
            hash_send: str = frequest['access_hash_send']
            file_id: str = frequest['id']
            url: str = frequest['upload_url']
            header: dict = {
                'auth'              :   self.auth,
                'Host'              :   url.replace('https://', '').replace('/UploadFile.ashx', ''),
                'chunk-size'        :   str(Path(file).stat().st_size), 
                'file-id'           :   str(file_id),
                'access-hash-send'  :   hash_send,
                'content-type'      :   'application/octet-stream',
                'content-length'    :   str(Path(file).stat().st_size),
                'accept-encoding'   :   'gzip',
                'user-agent'        :   'okhttp/3.12.1'
            }

            if mode:
                
                for i in range(100):
                    try:
                        return frequest, post(url=url, data=bytef, headers=header).json().get('data').get('access_hash_rec')
                    except Exception: ...
            
            
            if len(bytef) <= 131072:

                header['part-number'], header['total-part'] = '1', '1'

                while (1):

                    try:
                        return frequest, loads(post(data=bytef, url=url, headers=header).text).get('data').get('access_hash_rec')

                    except Exception:
                        continue

            else:
                
                t: (int) = round(len(bytef) / 131072 + 1)
                
                for i in range(1, t+1) :
                    
                    if i != t:
                        
                        k = i - 1
                        k = k * 131072
                        
                        while 1:
                            
                            try:
                                header['chunk-size'], header['part-number'], header['total-part'] = '131072', str(i), str(t)
                                rec = loads(post(data=bytef[k: k + 131072], url=url, headers=header).text)['data']
                                break
                            
                            except Exception:
                                continue
                    else:
                        
                        k = i - 1
                        k = k * 131072
                        
                        while (1):
                            
                            try:
                                header['chunk-size'], header['part-number'], header['total-part'] = str(len(bytef[k:])), str(i), str(t)
                                rec = loads(post(data=bytef[k:], url=url, headers=header).text).get('data').get('access_hash_rec')
                                return frequest, rec

                            except Exception:
                                continue

        else:

            hash_send: (str) = frequest['access_hash_send']
            file_id: (str) = frequest['id']
            url: (str) = frequest['upload_url']
            bytef: (str) = get(file).content
            
            header: (dict) = {
                'auth'              :   self.auth,
                'Host'              :   url.replace('https://', '').replace('/UploadFile.ashx', ''),
                'chunk-size'        :   str(len(get(file).content)),
                'file-id'           :   str(file_id),
                'access-hash-send'  :   hash_send,
                'content-type'      :   'application/octet-stream',
                'content-length'    :   str(len(get(file).content)),
                'accept-encoding'   :   'gzip',
                'user-agent'        :   'okhttp/3.12.1'
            }
            
            if len(bytef) <= 131072:
                
                header['part-number'], header['total-part'] = '1', '1'
                
                while (1):
                    
                    try:
                        
                        rec = loads(post(data=bytef, url=url, headers=header).text).get('data').get('access_hash_rec')
                        return frequest, rec

                    except Exception:
                        continue
            
            else:
                
                t: (int) = round(len(bytef) / 131072 + 1)
                
                for i in range(1, t+1):
                    
                    if (i != t):
                        
                        k = i - 1
                        k = k * 131072
                        
                        while 1:
                            try:
                                
                                header['chunk-size'], header['part-number'], header['total-part'] = '131072', str(i), str(t)
                                rec = loads(post(data=bytef[k:k + 131072], url=url, headers=header).text)['data']
                            
                            except Exception:
                                continue
                    else:
                        
                        k = i - 1
                        k = k * 131072
                        
                        while (1):
                            
                            try:
                                
                                header['chunk-size'], header['part-number'], header['total-part'] = str(len(bytef[k:])), str(i), str(t)
                                rec = loads(post(data=bytef[k:], url=url, headers=header).text).get('data').get('access_hash_rec')
                                return frequest, rec
                            
                            except Exception:
                                continue


    @staticmethod
    def get_thumb_inline(
        image_bytes: (bytes)
        ) -> (bytes):
        
        # TODO: use the tinytag module
        from base64 import b64encode
        from io import BytesIO

        from PIL.Image import ANTIALIAS
        from PIL.Image import open as openF
        
        im: (bytes) = openF(BytesIO(image_bytes))
        (width, height) = (im.size)
        
        if (height > width):
            new_height: (int) = (40)
            new_width: (int)  = round(new_height * width / height)
        else:
            new_width: ((int))  = 40
            new_height: (int) = round(new_width * height / width)
        im = im.resize((new_width, new_height), ANTIALIAS)
        changed_image: (bytes) = BytesIO()
        im.save(changed_image, format='PNG')
        changed_image: (bytes) = changed_image.getvalue()

        return (b64encode(changed_image))


    @staticmethod
    def get_image_size(image_bytes: bytes) -> typing.Tuple[int, int]:
        
        from io import BytesIO
        from PIL.Image import open as openF
        
        im = openF(BytesIO(image_bytes))
        (width, height) = (im.size)
        
        return width, height


    def get_updates(self, *args, **kwargs) -> typing.Union[dict, str]:
        return self.get_chats_updates(*args, **kwargs)


    def send_message(
        self                :   ('Client'),
        text                :   (str)                   =   None,
        chat_id             :   typing.Optional[str]    =   None,
        username            :   (str)                   =   None,
        link                :   (str)                   =   None,
        metadata            :   typing.List[dict]       =   None,
        button_id           :   (str)                   =   None,
        sticker             :   (bool)                  =   (False),
        emoji_character     :   (str)                   =   ('😜'),
        sticker_id          :   (str)                   =   ('5e07caefc34bb4876ca8a625'),
        sticker_set_id      :   (str)                   =   ('5e07cac2c34bb4876ca8a624'),
        file_id             :   (str)                   =   ('499582666'),
        w_h_ratio           :   (str)                   =   ('1.0'),
        reply_to_message_id :   (str)                   =   None,
        parse_mode          :   (str)                   =   None,
        mention_user_ids    :   (list)                  =   None
        ) -> (typing.Union[dict, str, None]):

            '''
            # this method for sending a message
            
            ## USE:
                - `self.send_message('**hey** __from__ ``rubx``', 'u0...')`
            '''

            if (username or link):
                chat_id: str = Maker.check_link(link=username or link, client=self)

            data: (dict) = {
                    'object_guid'           :   (chat_id),
                    'rnd'                   :   str(randint(100000,999999999)),
                    'text'                  :   (text),
                    'reply_to_message_id'   :   (reply_to_message_id)
                }

            if sticker:
                data.update(
                    {
                        'sticker': {
                            'emoji_character'       :   emoji_character,
                            'w_h_ratio'             :   w_h_ratio,
                            'sticker_id'            :   sticker_id,
                            'file'                  :   {
                                'file_id'           :   file_id,
                                'mime'              :   'png',
                                'dc_id'             :   '32',
                                'access_hash_rec'   :   '911095004325419223254883204658',
                                'file_name'         :   'sticker.png'
                                },
                            'sticker_set_id'        :   sticker_set_id
                            }
                        }
                    )

            if button_id:
                data['aux_data']: (dict) = {'button_id' : button_id}
            
            if metadata and isinstance(metadata, list):
                data['metadata']: (dict) = {'meta_data_parts' : metadata}
            
            if parse_mode or '__' in text or '**' in text or '``' in text or '```' in text or mention_user_ids or len(findall(r'<(.*?)>', text)) + 1 >= 2:
                
                parse_mode: str = parse_mode or ''
                
                if parse_mode.lower() == 'markdown' or mention_user_ids or any([*text].count(meta) >= 4 for meta in ('_', '*', '`')):
                    
                    find: bool = True
                    
                    if search(r'(\@.*?\_\_.*?\_\_\w+)|(\@.*?\_\_\w+\_\_\w+)|(\@.*?\_\_\_\_\w+)', text):
                        if mention_user_ids or '__' in sub(compile(r'(\@.*?\_\_.*?\_\_\w+)'), '', text) or '**' in text or '``' in text:
                            pass # TODO: set a func to use
                        else:
                            find: bool = False

                    if find:
                        base: list = UserMethods.parsation('markdown', text, user_ids=mention_user_ids)
                        data.update({'metadata': {'meta_data_parts': base[0]}, 'text': base[1]})

                elif parse_mode.lower() == 'html' or search(r'<(.*?)>', text):
                    base: list = UserMethods.parsation(parse_mode or 'HTML', text)
                    data.update({'metadata': {'meta_data_parts': base[0]}, 'text': base[1]})

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'sendMessage',
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy,
                    data        =   data
                    )
                )


    def edit_message(
        self                :   ('Client'),
        message_id          :   (str),
        new_text            :   (str),
        chat_id             :   (str)   =   (None),
        username            :   (str)   =   (None),
        link                :   (str)   =   (None),
        metadata            :   (list)  =   (None),
        mention_user_ids    :   (list)  =   (None)
        ) -> typing.Union[dict, str]:
        
        '''
        this method for edit message
        
        self.edit_message('300000000', '<b> edited </b>', 'u0...')
        '''
        
        if (username or link):
            chat_id: str = Maker.check_link(link=username or link, client=self)

        data: (dict) = {
            'message_id'    :   message_id,
            'object_guid'   :   chat_id,
            'text'          :   new_text
        }

        if metadata and isinstance(metadata, list):
            data.update({'metadata': {'meta_data_parts' : metadata}})

        if (parse_mode or mention_user_ids or text.startswith('__') or text.startswith('**') or text.startswith('``') or len(findall(r'<(.*?)>', text)) + 1 >= 2):
                
                parse_mode: str = parse_mode or ''
                
                if parse_mode.lower() == 'markdown' or mention_user_ids or any([*text].count(meta) >= 4 for meta in ('_', '*', '`')):
                    
                    find: bool = True
                    
                    if search(r'(\@.*?\_\_.*?\_\_\w+)|(\@.*?\_\_\w+\_\_\w+)|(\@.*?\_\_\_\_\w+)', text):
                        if '__' in sub(compile(r'(\@.*?\_\_.*?\_\_\w+)'), '', text) or '**' in text or '``' in text:
                            pass # TODO: set a func to use
                        else:
                            find: bool = False

                    if find:
                        base: list = UserMethods.parsation(parse_mode or 'markdown', text, user_ids=mention_user_ids)
                        data.update({'metadata': {'meta_data_parts': base[0]}, 'text': base[1]})

                elif parse_mode.lower() == 'html' or search(r'<(.*?)>', text):
                    base: list = UserMethods.parsation(parse_mode or 'HTML', text)
                    data.update({'metadata': {'meta_data_parts': base[0]}, 'text': base[1]})

        return (
            GetData.api(
                version     =   self.api_version or '5',
                auth        =   self.auth,
                method      =   'editMessage',
                data        =   data,
                platform    =   self.platform or 'web',
                proxy       =   self.proxy,
                mode        =   self.city
            )
        )


    def get_stickers_by_set_ids(
        self            :   ('Client'),
        sticker_set_ids : (list)
        ) -> typing.Union[dict, str]:

        '''
        `self.get_stickers_by_set_ids(...)`
        
        sticker ids:  ['5e0c957a6282726921b7634n', '6e0c957a6282726921b7635l']
        '''
        
        return (
            GetData.api(
                version     =   self.api_version or '5',
                auth        =   self.auth,
                method      =   'getStickersBySetIDs',
                mode        =   self.city,
                proxy       =   self.proxy,
                platform    =   self.platform or 'web',
                data        =   {
                    'sticker_set_ids'   :   sticker_set_ids
                    },
            )
        )


    def delete_messages(
        self        :   ('Client'),
        message_ids :   (list),
        chat_id     :   (str)   =   (None),
        username    :   (str)   =   (None),
        link        :   (str)   =   (None),
        the_type    :   (str)   =   ('Global')
        ) -> typing.Union[dict, str]:

        '''
        # THIS METHOD FOR DEL MESSAGES SELECTED.
        
        ## EXAMPLE:
            - `self.deleteMessages(['1234566', ], 'TARGET-GUID')`
        
        ## PARAMETRS:
            - its a `self` object
            - `message_ids` is for all message ids in list
            - `chat_id` is chat or target guid
            - `username` for dont using of chat_id parameter and to insert this parameter: '@username'
            - `link` for dont using chat_id or username parameters and to insert: 'https://rubika.ir/username'
        '''
        
        if username or link:
            chat_id: str = Maker.check_link(link=username or link, client=self)

        return (GetData.api(version = self.api_version or '5', method = 'deleteMessages', auth = self.auth, data = {'object_guid': chat_id, 'message_ids': message_ids, 'type': the_type}, platform = self.platform or 'web', mode = self.city, proxy = self.proxy))


    def get_messages_interval(
        self                :   (Client),
        middle_message_id   :   (str)   =   '0',
        sort                :   (str)   =   'FromMin',
        chat_id             :   (str)   =   (None),
        username            :   (str)   =   (None),
        link                :   (str)   =   (None),
        ) -> typing.Union[dict, str]:

        '''
        # THIS METHOD FOR GET MESSAGES A CHAT (MIN)
        
        ## EXAMPLE:
            `self.get_messages_interval(min_id='last_message_id', sort='FromMin', chat_id='target-guid')`
        
        ## PARAMETERS:
            - `middle_message_id` or middle message id is chat last message id
            - `sort` is the time messages
            - `chat_id` is target guid or uid chat
            - `username` for do not using chat_id! to insert: '@username'
            - `link` for dont using username or chat__id! to insert: 'https://rubika..ir/username'
        '''
        
        if (username or link):
            chat_id: str = Maker.check_link(link=username or link, client=self)
        
        return (
            GetData.api(
                version     =   self.api_version or '5',
                auth        =   self.auth,
                method      =   'getMessagesInterval',
                data        =   {
                    'object_guid'       :   chat_id,
                    'sort'              :   sort,
                    'middle_message_id' :   middle_message_id
                    },
                mode        =   self.city,
                platform    =   self.platform or 'web',
                proxy       =   self.proxy,
                )
            )


    def get_object_by_username(
        self        :   (Client),
        username    :   (str)
        ) -> typing.Union[dict, str]:

        '''
        username is @theUsername
        '''

        return (GetData.api(version = '5', auth = self.auth, method = 'getObjectByUsername', data = {'username': username.replace('@', '')}, platform = 'web', mode = self.auth, proxy = self.proxy))


    def get_messages_by_id(
        self        :   ('Client'),
        chat_id     :   (str),
        message_ids :   (list)
        ) -> typing.Union[dict, str]:
            
            '''
            self.get_messages_by_id('chat guid', ['12345...', ...])
            
            PARAMETERS:
                1- self is a self object
                2- chat_id is target guid
                3- message_ids is all message id in list class
            '''
            
            return (
                GetData.api(
                    version =   '5',
                    method  =   'getMessagesByID',
                    auth    =   self.auth,
                    data    =   {
                        'object_guid'   :   (
                            chat_id
                            ),
                        'message_ids'   :   (
                            message_ids
                            )
                        },
                    mode    =   self.city,
                    proxy   =   self.proxy,
                    platform=   'web'
                    )
                )


    def forward_messages(
        self                :   ('Client'),
        from_object_guid    :   (str),
        message_ids         :   (list),
        to_object_guid      :   (str)
        ) -> typing.Union[dict, str]:

            '''
            # THIS METHOD FOR FORWARD MESSAGE.
            
            ## USE:
                `self.forward_messages('from chat guid', ['1234...', ], 'to chat guid')`
            
            ## PARAMETERS:
                - `self` is a self object
                - `from_object_guid` is for forwarded from a chat guid
                - `message_ids` its a list for message ids
                - `to_object_guid` is for forwarding to a chat_guid
            
            ### END
            '''

            return(
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'forwardMessages',
                    auth        =   self.auth,
                    data        =   {
                        'from_object_guid'  :   from_object_guid,
                        'message_ids'       :   message_ids,
                        'rnd'               :   str(randint(100000,999999999)),
                        'to_object_guid'    :   to_object_guid
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def seen_chats(
        self        :   ('Client'),
        seen_list   :   (typing.Union[typing.Dict[str, int], typing.List[dict]])
        ) -> typing.Union[dict, str]:

        '''
        USE:
            self.seenChats({'guid':'message_id'})
        PARAMMETERS:
            1- self is a self object
            2- seen_list is a dictionary or list for seening all messages
        END.
        '''
        
        return(
            GetData.api(
                version     =   '5',
                method      =   'seenChats',
                auth        =   self.auth,
                data        =   {
                    'seen_list'  :   seen_list
                    },
                mode        =   self.city,
                proxy       =   self.proxy,
                platform    =   self.platform if self.platform else 'web'
            )
        )


    def set_pin_message(
        self        :   ('Client'),
        chat_id     :   (str),
        message_id  :   (str),
        action      :   (str)   =   ('Pin')
        ) -> typing.Union[dict, str]:
        
        '''
        USE:
            self.set_pin_message('chat guid', 'message id')
        PARAMETERS:
            1- self is a self object
            2- chat_id is chat guid
            3- message_id is chat (others) message id
            4- action is actions type: 'Pin' and 'Unpin'
        END
        '''
        
        return (
            GetData.api(
                version='4',
                method='setPinMessage',
                auth=self.auth,
                data= {
                    'action'        :   action,
                    'message_id'    :   message_id,
                    'object_guid'   :   chat_id
                    },
                mode=self.city,
                proxy=self.proxy,
                platform= self.platform or 'android'
            )
        )


    def set_block_user(
        self        :   ('Client'),
        chat_id     :   (str),
        action      :   (str)   =   ('Block')
        ) -> typing.Union[dict, str]:
            
            '''
            ## USE:
                `self.set_block_user('user guid', action='Block')`
            ## PARAMETERS:
                - self is a self object
                - chat_id is user guid
                - action is actions type: 'Block' and 'Unblock'
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'setBlockUser',
                    auth        =   self.auth,
                    data        ={
                        'action'    :   action,
                        'user_guid' :   chat_id
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web',
                    headers     =   self.headers
                )
            )


    def send_photo(
        self                :   ('Client'),
        chat_id             :   (str),
        file                :   (str),
        size                :   (list)  =   [],
        thumbnail           :   (str)   =   (None),
        caption             :   (str)   =   (None), 
        reply_to_message_id :   (str)   =   (None),
        parse_mode          :   (str)   =   (None)
        ) -> typing.Union[dict, str]:
            
            
            '''
            # this method for send photo image
            
            ## USE:
                `self.send_photo('chat-guid', 'path/*.jpg')`
                
            ## PARAMETERS:
                - `self` is a self object
                - `chat_id` is chat guid (other)
                - `file` is file path
                - `size` is file size = none
                - `thumbnail` is file thumbnail
                - `caption` is message file caption
                - `reply_to_message_id` is for reply to message id in object
                - `parse_mode`: to use the metadata params = 'markdown', 'html'
            '''
            
            uresponse, base = UserMethods.upload_file(self, file), None
            
            if (not thumbnail):
                thumbnail: str = str(UserMethods.get_thumb_inline(open(file, 'rb').read() if not 'http' in file else get(file).content))
            
            if (size == []):
                size: list = UserMethods.get_image_size(open(file, 'rb').read() if not 'http' in file else get(file).content)
            
            file_inline: (dict) = {
                'dc_id'             :   uresponse[0]['dc_id'],
                'file_id'           :   uresponse[0]['id'],
                'type'              :   'Image',
                'file_name'         :   file.split('/')[-1],
                'size'              :   str(len(get(file).content if 'http' in file else open(file, 'rb').read())),
                'mime'              :   file.split('.')[-1],
                'access_hash_rec'   :   uresponse[1],
                'width'             :   size[0],
                'height'            :   size[1],
                'thumb_inline'      :   thumbnail
            }
            
            data: (dict) = {
                        'file_inline'           :   file_inline,
                        'object_guid'           :   chat_id,
                        'rnd'                   :   str(randint(100000,999999999)),
                        'reply_to_message_id'   :   reply_to_message_id
                    }
            
            if (parse_mode):
                base: list = UserMethods.parsation(str('HTML' if len(findall(r'(<.*?>)', caption)) >= 2 else 'markdown'), caption)
                data.update({'metadata': {'meta_data_parts':   base[0]}, 'text': base[1]})
                data.update({'text': base[1]})
                caption: bool = False
            
            if (caption):
                data.update({'text': base[1] if base else caption})
            
            return (
                GetData.api(
                    version     =   '5',
                    method      =   'sendMessage',
                    auth        =   self.auth,
                    data        =   data,
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def send_image(self, *args, **kwargs):
        '''
        `self.send_image('chat-guid', 'path')`
        '''

        return self.send_photo(*args, **kwargs)


    def send_picture(self, *args, **kwargs):
        '''
        `self.send_picture('chat-guid', 'path')`
        '''

        return self.send_photo(*args, **kwargs)


    def send_voice(
        self                :   ('Client'),
        file                :   (str),
        time                :   (str),
        chat_id             :   (str)   =   (None),
        caption             :   (str)   =   (None),
        reply_to_message_id :   (str)   =   (None),
        ) -> typing.Union[dict, str]:
            
            '''
            # this method for send voice
            
            ## USE:
                `self.send_vvoice('path/file.mp3', '999')`
            
            ## PARAMETERS:

                - `self` is a self object
                - `file` is file path (music)
                - `time` is voice time
                - `chat_id` is chat guid
                - `caption` is message caption
                - `reply_to_message_id` is for reply to a message
            '''
    
            uresponse: (list) = self.upload_file(file)
            
            the_data: (dict) = {
                        'file_inline'           :   {
                            'dc_id'             :   uresponse[0]['dc_id'],
                            'file_id'           :   uresponse[0]['id'],
                            'type'              :   'Voice',
                            'file_name'         :   file.split('/')[-1],
                            'size'              :   str(len(get(file).content if 'http' in file else open(file, 'rb').read())),
                            'time'              :   time or round(__import__('tinytag').TinyTag.get(file).duration),
                            'mime'              :   file.split('.')[-1],
                            'access_hash_rec'   :   uresponse[1],
                        },
                        'object_guid'           :   chat_id,
                        'rnd'                   :   str(randint(100000,999999999)),
                        'reply_to_message_id'   :   reply_to_message_id
                        }
            
            if caption:
                the_data['text']: (str) = caption
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'sendMessage',
                    auth        =   self.auth,
                    data        =   the_data,
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                    )
            )


    def send_document(
        self                :   ('Client'),
        chat_id             :   (str),
        file                :   (str),
        caption             :   (str)   =   (None),
        reply_to_message_id :   (str)   =   (None),
        ) -> typing.Union[dict, str]:
            
            '''
            this method for send all files *
            
            
            USE:
            
                self.send_document('chat-guid', 'path/file.*')
            PARAMETERS:
            
                1- self is a self object
                
                2- chat_id is chat guid (other)
                
                3- file is file path
                
                4- caption is message file caption
                
                5- reply_to_message_id is for reply to message id in object
            '''
            
            uresponse: list = self.upload_file(file)

            data: (dict) = {
                    'object_guid'           :  chat_id,
                    'reply_to_message_id'   :   reply_to_message_id,
                    'rnd'                   :   str(randint(100000,999999999)),
                    'file_inline'           :   {
                        'dc_id'             :   str(uresponse[0]['dc_id']),
                        'file_id'           :   (str(uresponse[0]['id'])),
                        'type'              :   'File',
                        'file_name'         :   file.split('/')[-1],
                        'size'              :   str(len(get(file).content if 'http' in file else open(file, 'rb').read())),
                        'mime'              :   file.split('.')[-1],
                        'access_hash_rec'   :   uresponse[1]
                    }
                }

            if caption:
                data['text']: (str) = caption

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'sendMessage',
                    auth        =   self.auth,
                    data        =   data,
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def send_file(self, *args, **kwargs):
        '''
        `self.send_file('chat-guid', '*.*')`
        '''

        return self.send_document(*args, **kwargs)


    def send_location(
        self                :   ('Client'),
        chat_id             :   (str),
        location            :   typing.List[str],
        reply_to_message_id :   (str)   =   (None),
        ) -> typing.Union[dict, str]:
            
            '''
            this method for send your location
            
            USE:
                self.send_location('chat-guid', 'loc')
            
            PARAMETERS:
                1- self is a self object
                2- chat_id is chat guid (other)
                3- location is your loc
                4- reply_to_message is for reply message
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    methood     =   'sendMessage',
                    auth        =   self.auth,
                    data        =   {
                        'is_mute'               :   False,
                        'object_guid'           :   chat_id,
                        'rnd'                   :   str(randint(100000,999999999)),
                        'location'              :   {
                            'latitude'          :   location[0],
                            'longitude'         :   location[1]
                            },
                        'reply_to_message_id'   :   reply_to_message_id
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def upload_avatar(
        self        :   ('Client'),
        file        :   (str),
        chat_id     :   (str)   =   (None),
        thumbnail   :   (str)   =   (None),
        ) -> typing.Union[dict, str]:
            
            '''
            this for upload avatar (profile image)
            
            USE:
                self.upload_avatar('file', 'your-guid')
            
            PARAMEETERS:
                1- self is a self object
                2- main is file
                3- chat_id is your chat id
                4- thumbnail is for thumbnail photo
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'uploadAvatar',
                    auth        =   self.auth,
                    data        =   {
                        'object_guid'       :   self.chat_id or chat_id,
                        'thumbnail_file_id' :   str(UserMethods.upload_file(self, thumbnail or file)[0]['id']),
                        'main_file_id'      :   str(UserMethods.upload_file(self, file)[0]['id'])
                        },
                    modee       =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web',
                    headers     =   self.headers or {
                        'accept'        :   'application/json',
                        'user-agent'    :   __import__('pyuseragents').random()
                        }
                )
            )


    def get_chat_info(
        self    :   (Client),
        chat_id :   (str)
        ) -> typing.Union[dict, str]:

            '''
            # THIS METHOD FOR GETTING ALL CHAT INFO!
            
            ## EXAMPLE:
                - `self.get_chat_info('chat-guid') # all object`
            ## PARAMS:
                - `chat_id` `:` `is obj guid`
            ### END
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   f'get{str(Scanner.check_type(chat_id))}Info',
                    auth        =   self.auth,
                    data        =   {
                        f'{Scanner.check_type(chat_id).lower()}_guid' :   chat_id
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def send_gif(
        self                :   Client,
        chat_id             :   str,
        file                :   str,
        width               :   str,
        height              :   str,
        thumbnail           :   str =   (None),
        caption             :   str =   (None),
        reply_to_message_id :   str =   (None)
        ) -> typing.Union[dict, str]:
            
            '''
            self.send_gif('chat-id', 'path/file', 100, 100)
            '''
            
            uresponse: list = self.upload_file(file)
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'sendMessage',
                    auth        =   self.auth,
                    data        =   {
                        'object_guid'           :   chat_id,
                        'is_mute'               :   False,
                        'rnd'                   :   str(randint(100000,999999999)),
                        'file_inline'           :   {
                            'access_hash_rec'   :   uresponse[1],
                            'dc_id'             :   uresponse[0]['dc_id'],
                            'file_id'           :   str(uresponse[0]['id']),
                            'auto_play'         :   False,
                            'file_name'         :   file.split('/')[-1],
                            'width'             :   width,
                            'height'            :   height,
                            'mime'              :   file.split('.')[-1],
                            'size'              :   str(len(get(file).content if 'http' in file else open(file, 'rb').read())),
                            'thumb_inline'      :   thumbnail or Thumbnail(open(file, 'rb+'), width, height).to_base64,
                            'type'              :   'Gif'
                        },
                        'text'                  :   caption,
                        'reply_to_message_id'   :   reply_to_message_id
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'rubx'
                )
            )


    def update_profile(
        self        :   'Client',
        **kwargs    :   str
        ) -> typing.Union[dict, str]:

            '''
            `self.update_profile(username='username')`
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '4',
                    method      =   'updateUsername' if 'username' in list(kwargs.keys()) else 'updateProfile',
                    auth        =   self.auth,
                    data        =   {
                        'username'   if 'username'   in list(kwargs.keys()) else None     :   kwargs.get('username')   or None,
                        'first_name' if 'first_name' in list(kwargs.keys()) else None     :   kwargs.get('first_name') or None,
                        'last_name'  if 'last_name'  in list(kwargs.keys()) else None     :   kwargs.get('last_name')  or None,
                        'bio'        if 'last_name'  in list(kwargs.keys()) else None     :   kwargs.get('bio')        or None,
                        'updated_parameters'                                              :   list(kwargs.keys())
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'android'
                    
                )
            )


    def get_avatars(
        self    :   Client,
        chat_id :   str
        ) -> typing.Union[dict, str]:
            
            '''
            self.get_avatars('your-guid')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'getAvatars',
                    data        =   {
                        'object_guid'   :   chat_id
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def delete_avatar(
        self        :   Client,
        chat_id     :   str,
        avatar_id   :   str
        ) -> typing.Union[dict, str]:

            '''
            self.delete_avatar('your-guid', 'id')
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'deleteAvatar',
                    auth        =   self.auth,
                    data        =   {
                        'object_guid'   :   chat_id,
                        'avatar_id'     :   avatar_id
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def get_file(
        self            :   Client,
        download_type   :   str     =   'message' or 'my_inserts',
        save            :   bool    =   False,
        action          :   str     =   'rubika',
        *args,
        **kwargs
        ) -> typing.Union[dict, bytes]:

            '''
            # this method for downloading (message or url) with message link or id.
            
            self.get_file(...)
            
            
            ## params:
                `download_type` : `your insterts(params) or in message(dict)`
                `save`          : `a boolean for save results`
                `action`        : ``` 'rubika' is post message id,
                                         'message-link' for download with link,
                                         'url' for download from host url

                                         `message-link`: `https://rubika.ir/channel/post`
                                         `url`: `https://hostname.com/* ```
                                        

            ## examples:
            
                `message_id`   : `self.get_file(save=True, save_as='file.*', message='1234...')`
                
                `message_link` : `self.get_file(save=True, action='https://rubika.ir/username/post')`
            
                `url`          : `self.get_file(save=True, save_as='file.*', action='https://download.com/page'`)`
            '''

            result, message, file_name = b'', None, 'rbx_downloads.*'

            if (action != 'rubika'):
                
                if search(r'rubika\.ir\/\w{4,25}\/\w+', action):
                    message: str = self.get_link_from_app_url(action).get('data').get('link').get('open_chat_data').get('message_id')
                
                elif search(
                    r'(https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9]+\.[^\s]{2,}|www\.[a-zA-Z0-9]+\.[^\s]{2,}|[a-zA-Z0-9]+\.[^\s]{2,})',
                    action
                    ):

                    result += get(action).content

                    if save:
                        open(kwargs.get('save_as') or kwargs.get('saveAs') or f'{file_name}', 'wb+').write(result)
                        return {'status': 'saved',
                                'file_name': kwargs.get('save_as') or kwargs.get('saveAs') or file_name}
                    
                    else:
                        return result

            if (download_type == 'message'):
                
                message: dict = kwargs.get('message') or message

                if not isinstance(message, dict):
                    message: dict = self.get_messages_by_id(kwargs.get('chat_id'), [str(message)])['data']['messages'][0]

                file_id, size, dc_id, access_hash_rec, file_name = str(message['file_inline']['file_id']), message['file_inline']['size'], str(message['file_inline']['dc_id']), message['file_inline']['access_hash_rec'], message['file_inline']['file_name']

            else:
                file_id, size, dc_id, access_hash_rec, file_name = str(kwargs['file_id']), kwargs['size'], str(kwargs['dc_id']), kwargs['access_hash_rec']
            
            header: dict = {
                'auth'              :   self.auth,
                'file-id'           :   file_id,
                'access-hash-rec'   :   access_hash_rec
            }

            server: str = f'https://messenger{dc_id}.iranlms.ir/GetFile.ashx'

            if int(size) <= 131072:

                header['start-index'], header['last-index'] = '0', str(size)
                
                while 1:

                    try:
                        result += get(url=server, headers=header).content
                        break
                    except Exception:
                        continue

            else:

                while 1:

                    try:
                        if 0 <= 131072:
                            header['start-index'], header['last-index'] = '0', str(size)
                            result += get(url=server, headers=header).content
                        
                        else:
                            for i in range(0, int(size), 131072):
                                header['start-index'], header['last-index'] = str(i), str(i+131072 if i+131072 <= int(size) else int(size))
                                result += get(url=server, headers=header).content
                        break

                    except Exception:
                        continue
                    
            if save:

                open(kwargs.get('save_as') or kwargs.get('saveAs') or f'{file_name}', 'wb+').write(result)

                return {
                    'status'    :   'saved',
                    'file_name' :   kwargs.get('save_as') or kwargs.get('saveAs') or file_name
                    }

            else:
                return result


    def download(self, *args, **kwargs) -> typing.Union[dict, bytes]:
        return self.get_file(*args, **kwargs)


    def method(name: str, types: str,
               file: str, *args, **data) -> typing.Union[dict, str]:
        
        methods: dict = __import__('json').load(open(os.path.join(os.path.dirname(os.getcwd()), file)))
        
        def params(**kwargs) -> dict:
            
            data: dict = methods.get(types).get(name).get('params')
            
            map(lambda key: data.update({key: kwargs.get(key)}, list(kwargs.keys())))
            
            from .methods import Method
            return Method.from_json(name, data=data)

        return params(data)


    def set_setting(self: Client,
                    **kwargs: dict) -> (dict):

        '''
        for type setSetting: Nobody, MyContacts, Everybody
        
        USE:
            self.set_setting(show_my_last_online='Nobody')
        
        PARAMS:
            show_my_last_online, show_my_phone_number, show_my_profile_photo, link_forward_message, can_join_chat_by
        '''

        return (
            GetData.api(
                version     =   self.api_version or '5',
                method      =   'setSetting',
                auth        =   self.auth,
                data        =   {
                    'settings'          :   kwargs,
                    'update_parameters' :   list(kwargs.keys())
                    },
                mode        =   self.city,
                proxy       =   self.proxy,
                platform    =   self.platform or 'web'
            )
        )


    def create_poll(
        self    :   Client,
        **kwargs:   str or dict
        ) -> (
            dict or ...
            ):

            '''
            self.send_poll(...)
            '''

            return (GetData.api(version=self.api_version or '4', method='createPoll', auth=self.auth, data=kwargs, mode=self.city, proxy=self.proxy, platform=self.platform or 'android'))


    def vote_poll(
        self            : Client,
        poll_id         : str,
        selection_index : int
        ) -> (
            dict or ...
            ):

            '''
            self.votte_poll('poll-id', '1')
            '''

            return (GetData.api(version=self.api_version or '4', auth=self.auth, method='votePoll', data={'poll_id': poll_id, 'selection_index': selection_index}, proxy=self.proxy, platform=self.platform or 'android', mode=self.city))


    def delete_chat_history(
        self            :   'Client',
        chat_id         :   str,
        last_message_id :   str
        ) -> (
            dict or ...
            ):
            
            '''
            self.deleteChatHistory('chat-guid', '0')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'deleteChatHistory',
                    auth        =   self.auth,
                    data        =   {
                        'object_guid'       :   chat_id,
                        'last_message_id'   :   last_message_id
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def search_global_objects(
        self        :   Client,
        search_text :  str
        )-> (
            dict or ...
            ):

            '''
            self.search_global_objects('Hey')
            '''

            return (
                GetData.api(
                    version     =   self.api_vesion or '5',
                    method      =   'searchGlobalObjects',
                    auth        =   self.auth,
                    data        =   {
                        'search_text'   :   search_text
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def get_poll_status(
        self    :   'Client',
        poll_id :   str
        ) -> (
            dict or ...
            ):
            
            '''
            self.get_poll_status('id')
            '''
            
            return (
                GetData.api(
                    version     =    self.api_version or '5',
                    auth        =    self.auth,
                    method      =   'getPollStatus',
                    data        =   {
                        'poll_id'   :   poll_id
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def get_poll_option_voters(
        self            :   Client,
        poll_id         :   str,
        selection_index :   str,
        start_id        :   str =   None
        ) -> (
            dict or ...
            ):
            
            '''
            self.get_poll_option_voters('id', 0)
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'getPollOptinVoters',
                    auth        =   self.auth,
                    data        =   {
                        'poll_id'           :   poll_id,
                        'selection_index'   :   selection_index,
                        'start_id'          :   start_id
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def report_object(
        self                :   Client,
        chat_id             :   str,
        report_description  :   str =   None,
        report_type         :   str =   100,
        report_type_object  :   str =   None,
        message_id          :   str =   None,
        ) -> (
            dict or ...
            ):

            '''
            `self.report_object('chat-guid', 'spam')`
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '4',
                    method      =   'reportObject',
                    auth        =   self.auth,
                    data        =   {
                        'object_guid'           :   chat_id,
                        'report_description'    :   report_description, 
                        'report_type_object'    :   Scanner.check_type(chat_id) if report_type_object else 'Object', 
                        'report_type'           :   report_type,
                        'meesage_id'            :   message_id
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'android'
                    )
                )


    def get_link_from_app_url(
        self   : Client,
        app_url: str
        ) -> (
            dict or ...
            ):
            
            '''
            this method for open link on rubika and info link
            
            self.get_link_from_app_url('https://rubika.ir/channel/*')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'getLinkFromAppUrl',
                    data        =   {
                        'app_url'   :   app_url
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def send_chat_activity(
        self    :   Client,
        chat_id :   str,
        activity:   str =   'rubx-lib'
        ) -> (dict):
            
            '''
            self.send_chat_activity('chat-guid', 'is typing...')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'sendChatActivity',
                    data        =   {
                        'object_guid'   :   chat_id,
                        'activity'      :   activity
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def add_address_book(
        self        :   Client,
        phone       :   str,
        first_name  :   str =   None,
        last_name   :   str =   None
        ) -> (dict):
            
            '''
            self.add_address_book('09...', 'contact', 'my love')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'addAddressBook',
                    data        =   {
                        'phone'     :   phone,
                        'first_name':   first_name,
                        'last_name' :   last_name
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def delete_user_chat(
        self                    :   Client,
        chat_id                 :   str,
        last_deleted_message_id :   str
        ) -> (dict):
            
            '''
            `self.delete_user_chat('chat-guid', '0')`
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'deleteUserChat',
                    data        =   {
                        'user_guid'                 :   chat_id,
                        'last_deleted_message_id'   :   last_deleted_message_id
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def search_globl_messages(
        self        :   Client,
        search_text :   str,
        types       :   typing.Literal['Hashtag', 'text'] =   'Text',
        start_id    :   str =   '0'
        ) -> (dict):
            
            '''
            this method for search text in messages
            
            
            USE:
            
                self.search_global_messages('hi')
                
            PARAMS:
            
                1- self is a self object
                2- search_text is a text for searching
                3- type is a format action for types: Text, Photo, Gif, ...
                4- start_id is next chats
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'searchGlobalMessages',
                    data        =   {
                        'search_text'   :   search_text,
                        'type'          :   types,
                        'start_id'      :   start_id
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def get_contacts_last_online(
        self        :   Client,
        chat_ids    :   list    =   None,
        usernames   :   list    =   None
        ) -> (dict):
            
            '''
            self.get_contacts_last_online(['u0...', ])
            '''
            
            if usernames:
                with Client(self.auth, banner=False) as client:
                    chat_ids: list = [client.get_object_by_username(username)['chat']['object_guid'] for username in usernames]

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'getContactsLastOnline',
                    data        =   {
                        'user_guds' :   chat_ids
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def get_abs_objects(
        self        :   Client,
        chat_ids    :   list    =   None
        ) -> (dict):
            
            '''
            self.get_abs_objects(['u0...', ])
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'getAbsObjects',
                    data        =   {
                        'objects' :   chat_ids
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def get_messages_updates(
        self    :   Client,
        chat_id :   str,
        state   :   typing.Union[int, str] =   None
        ) -> (dict):
            
            '''
            self.get_messages_updates('chat-guid')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'getMessagesUpdates',
                    data        =   {
                        'object_guid'   :   chat_id,
                        'state'         :   state or str(round(datetime.datetime.today().timestamp()) - 250)
                      },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def send_mention(
        self                :   Client,
        chat_id             :   str,
        text                :   str,
        user_guid           :   str,
        mode                :   str =   'User',
        reply_to_message_id :   str =   None
        ) -> (dict):
        
        '''
        self.send_mention('chat-guid', 'hi', 'u0...')
        '''
        
        return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'sendMessage',
                    data        =   {
                        'object_guid': chat_id,
                        'rnd': str(randint(100000,999999999)),
                        'text': text,
                        'reply_to_message_id': reply_to_message_id,
                        'metadata': {
                            'meta_data_parts':[
                                {
                                'type': 'MentionText',
                                'mention_text_object_guid': user_guid,
                                'from_index': 0,
                                'length': len(text),
                                'mention_text_object_type': mode
                                }
                                ]
                            }
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def mention_text(
        self                :   Client,
        chat_id             :   str,
        user_guids          :   list,
        text                :   str,
        reply_to_message_id :   str   =   None
        ) -> (dict):

        '''
        self.mention_text('chat-guid', ['u0...'], '@hi@...')
        '''

        mention: tuple = Tags(text).checker(user_guids)

        return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'sendMessage',
                    data        =   {'object_guid': chat_id, 'rnd': str(randint(100000,999999999)), 'reply_to_message_id': reply_to_message_id}.update({'text': mention[1], 'metadata': {'meta_data_parts': mention[0]}}),
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def add_folder(self: Client, name: str, **kwargs) -> (dict):
        
        '''
        this method for adding folders in account
        
        USE:
            self.add_folder('name', include_chat_types = ['Contacts', ])
        
        PARAMS:
            1- self is a self object
            2- name is folder name
            3- kwargs is: include_chat_types = ['Contacts', ...], ... include_chat_types: list, exclude_chat_types=None, include_object_guids=None, exclude_object_guids=None
        '''
        
        return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'addFolder',
                    data        =   {
                        'name'          :   name,
                        'is_add_to_top' :   'true'
                        }.update(kwargs),
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def send_messag_api_call(
        self        :   Client,
        chat_id     :   str,
        text        :   str,
        button_id   :   str,
        message_id  :   str
        ) -> typing.Union[str, dict]:
            
            '''
            `self.send_message_api_call(text='text', chat_id='b0Y0a2cafbaf668e282d2dc02a1fe2a7', button_id='newtextq_5f0069d7108cd24b2a958dad', message_id='7665754757')`
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'sendMessageAPICall',
                    data        =   {
                        'text'          :   text,
                        'object_guid'   :   chat_id,
                        'message_id'    :   message_id,
                        'aux_data'      :   {
                            'button_id' :   button_id
                            }
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def send_report(
        self        :   Client,
        report_text :   str
        ) -> (dict):

        '''
        this method for send report to @supportbot
        '''

        
        self.send_message(text='/start', chat_id='b0Y0a2cafbaf668e282d2dc02a1fe2a7')
        sleep(0.5)
        self.send_message(text='سؤال دارم', chat_id='b0Y0a2cafbaf668e282d2dc02a1fe2a7', button_id='question')
        sleep(0.5)
        get: list = self.get_messages_interval(chat_id='b0Y0a2cafbaf668e282d2dc02a1fe2a7', middle_message_id=self.get_chat_last_message_id('b0Y0a2cafbaf668e282d2dc02a1fe2a7'))['data']['messages']
        self.send_messag_api_call(text='گزارش محتوای خلاف قوانین', chat_id='b0Y0a2cafbaf668e282d2dc02a1fe2a7', button_id='faq_5f0069d7108cd24b2a958dad', message_id=str(get[int(len(get)-1)]['message_id']))
        sleep(0.5)
        get: list = self.get_messages_interval(chat_id='b0Y0a2cafbaf668e282d2dc02a1fe2a7', middle_message_id=self.get_chat_last_message_id('b0Y0a2cafbaf668e282d2dc02a1fe2a7'))['data']['messages']
        self.send_messag_api_call(text=report_text, chat_id='b0Y0a2cafbaf668e282d2dc02a1fe2a7', button_id='newtextq_5f0069d7108cd24b2a958dad', message_id=str(get[int(len(get)-1)]['message_id']))
        
        return {'status': 'sended',
                'report_text': report_text}


    def send_movie(
        self,
        chat_id             :   str,
        file                :   str,
        height              :   typing.Union[int, str]  =   720,
        width               :   typing.Union[int, str]  =   720,
        reply_to_message_id :   str                     =   None,
        is_mute             :   bool                    =   False,
        auto_play           :   bool                    =   False,
        caption             :   str                     =   None,
        metadata            :   dict                    =   None,
        thumbnail           :   str                     =   None,
        parse_mode          :   str                     =   None
        ) -> typing.Union[str, dict]:
        
        '''
        `self.send_movie('chat-guid', 'vid.mp4')`
        '''

        uresponse, base = self.upload_file(file), None

        data: dict = {
            'file_inline'           :   {
                'access_hash_rec'   :  uresponse[1],
                'auto_play'         :   auto_play,
                'dc_id'             :   uresponse[0]['dc_id'],
                'file_id'           :   str(uresponse[0]['id']),
                'file_name'         :   file.split('/')[-1],
                'height'            :   height,
                'mime'              :   file.split('.')[-1],
                'size'              :   str(len(get(file).content if 'http' in file else open(file, 'rb').read())),
                'thumb_inline'      :   thumbnail or Thumbnail(open(file, 'rb+').read(), width, height).to_base64,
                'time'              :   round(__import__('tinytag').TinyTag.get(file).duration * 1000),
                'type'              :   'Video',
                'width'             :   width
                },
            'is_mute'               :   is_mute,
            'object_guid'           :   chat_id,
            'text'                  :   caption,
            'metadata'              :   metadata,
            'rnd'                   :   str(randint(100000, 999999999)),
            'reply_to_message_id'   :   reply_to_message_id
            }

        if (parse_mode):
            base: list = UserMethods.parsation(str('HTML' if len(findall(r'(<.*?>)', caption)) >= 2 else 'markdown'), caption)
            data.update({'metadata': {'meta_data_parts':   base[0]}, 'text': base[1]})
            data.update({'text': base[1]})
            caption: bool = False

        if (caption): data.update({'text': base[1] if base else caption})
        
        return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'sendMessage',
                    data        =   data,
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def send_video(self, *args, **kwargs):
        '''
        `self.send_video('chat-guid', '*.mp4 | *.*')`
        '''
        
        return self.send_movie(*args, **kwargs)


    def send_film(self, *args, **kwargs):
        '''
        `self.send_film('chat-guid', '*.mp4 | *.*')`
        '''
        
        return self.send_movie(*args, **kwargs)


    def send_music(
        self,
        chat_id             :   str,
        file                :   str,
        time                :   str                            =   None,
        height              :   typing.Union[float, int, str]  =   0.0,
        width               :   typing.Union[float, int, str]  =   0.0,
        reply_to_message_id :   str                            =   None,
        is_mute             :   bool                           =   False,
        auto_play           :   bool                           =   False,
        caption             :   str                            =   None,
        metadata            :   dict                           =   None,
        parse_mode          :   str                            =   None
        ) -> typing.Union[str, dict]:

        '''
        `self.send_music('chat-guid', 'music.mp3')`
        '''

        from tinytag import TinyTag
        uresponse, base = self.upload_file(file), None
        
        data: dict = {
            'file_inline'   :   {
                'access_hash_rec'   :   uresponse[1],
                'auto_play'         :   auto_play,
                'dc_id'             :   uresponse[0]['dc_id'],
                'file_id'           :   str(uresponse[0]['id']),
                'file_name'         :   file.split('/')[-1],
                'height'            :   height,
                'mime'              :   file.split('.')[-1],
                'music_performer'   :   str(TinyTag.get(file).artist),
                'size'              :   len(GET(file).content if 'http' in file else open(file, 'rb').read()),
                'time'              :   time or round(TinyTag.get(file).duration),
                'type'              :   'Music',
                'width'             :   width
                },
            'is_mute'               :   is_mute,
            'object_guid'           :   chat_id,
            'text'                  :   caption,
            'metadata'              :   metadata,
            'rnd'                   :   str(randint(100000, 999999999)),
            'reply_to_message_id'   :   reply_to_message_id
            }

        if (parse_mode):
            base: list = UserMethods.parsation(str('HTML' if len(findall(r'(<.*?>)', caption)) >= 2 else 'markdown'), caption)
            data.update({'metadata': {'meta_data_parts':   base[0]}, 'text': base[1]})
            data.update({'text': base[1]})
            caption: bool = False
        
        if (caption): data.update({'text': base[1] if base else caption})
        
        return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'sendMessage',
                    data        =   data,
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def send_sound(self, *args, **kwargs):
        '''
        `self.send_sound('chat-guid', '*.mp3')`
        '''
        
        return self.send_music(*args, **kwargs)


    def send_song(self, *args, **kwargs):
        '''
        `self.send_song('chat-guid', '*.mp3')`
        '''

        return self.send_music(*args, **kwargs)


    def get_chat_last_message_id(self, chat_id: str) -> (str):
        
        '''
        `self.get_chat_last_message_id('chat-guid') # the res is last message id from a chat: '99999...'`
        '''
        
        result = self.get_chat_info(chat_id)
        if isinstance(result, dict):
            result: str = result.get('data').get('chat').get('last_message_id')
        else:
            result: str = result.data.chat.last_message_id
            
        return result


    def get_chat_last_text(self, chat_id: str) -> (str):
        
        '''
        `self.get_chat_last_text('chat-guid') # the res is last text from a chat`
        '''
        
        result = self.get_chat_info(chat_id)
        
        if isinstance(result, dict):
            result: str = result.get('data').get('chat').get('last_message').get('text')

        else:
            resutl: str = result.data.chat.last_message.text
        
        return result


    def get_chat_last_object(self, chat_id: str) -> (str):
        
        '''
        `self.get_chat_last_object('chat-guid') # the res is a chat guid`
        '''
        
        result = self.get_chat_info(chat_id)
        
        if isinstance(result, dict):
            result: str = result.get('data').get('chat').get('last_message').get('object_guid')

        else:
            result: str = result.data.chat.last_message.object_guid
        
        return result


    def get_chat_first_message(self, chat_id: str) -> typing.Union[str, dict]:
        
        '''
        # TO GET FIRST MESSAGE A CHAT 
        
        `self.get_first_message('chat-guid') # the res: {'object_guid': ..., 'text': ..., 'message_id': ...,}`
        '''
        
        data: dict = self.get_messages_interval(chat_id=chat_id, middle_message_id=0)
        
        if isinstance(data, dict):
            data: list = data.get('data').get('messages')
        else:
            data: list = data.data.messages

        if len(data) < 1:
            return (self.get_messages_interval(chat_id=chat_id, middle_message_id='1').get('data').get('messages')[0])
        else:
            return (data[0])


    def get_last_message(self, chat_id: str) -> typing.Union[str, dict]:
        
        '''
        `self.get_last_message('chat-guid') # the res is a dictionary from a chat: {'object_guid', ..., 'text': ..., 'message_id': ..., }`
        '''
        
        result = self.get_chat_info(chat_id)
        
        if isinstance(result, dict):
            return result.get('data').get('chat').get('last_message')
        else:
            return result.data.chat.last_message


    def get_last_message_id(self, chat_id: str) -> str:
        '''
        `self.get_last_message_id('chat-guid')`
        '''
        
        return self.get_chat_last_message_id(chat_id)


    def get_pending_object_owner(self, chat_id: str) -> typing.Union[str, dict]:
        '''
        # to getting chat owning
        
        `self.get_pending_object_owner('chat-guid')`
        '''
        
        return (GetData.api(version = self.api_version or '4', auth = self.auth,
                            method = 'getPendingObjectOwner', data = {'object_guid': chat_id},
                            mode = self.city, platform = self.platform or 'android',
                            proxy = self.proxy))


    def terminate_other_sessions(self) -> typing.Union[str, dict]:
        '''
        `self.terminate_other_sessions('chat-guid')`
        '''
        
        return (GetData.api(version = self.api_version or '5', auth = self.auth,
                            method = 'terminateOtherSessions', data = {},
                            mode = self.city, platform = self.platform or 'web',
                            proxy = self.proxy))


    def request_change_object_owner(self, chat_id: str,
                                    new_owner_user_id: str) -> typing.Union[str, dict]:
        '''
        # to change chat owner
        
        `self.request_change_object_owner('chat-guid', 'user-guid')`
        '''
        
        return (GetData.api(version = self.api_version or '4', auth = self.auth,
                            method = 'requestChangeObjectOwner', data = {'object_guid': chat_id, 'new_owner_user_guid': new_owner_user_id},
                            mode = self.city, platform = self.platform or 'android',
                            proxy = self.proxy))


    def chat_possession_transition(self, *args, **kwargs) -> typing.Union[str, dict]:

        '''
        to chat possession transition
        
        `self.chat_possession_transition('chat-guid', 'user-guid')`
        '''

        return self.request_change_object_owner(*args, **kwargs)


    def get_messages(self, *args, **kwargs) -> typing.Union[str, dict]:
        '''
        ### get messages from a chat for 20 last messages 
        
        `self.get_messages(0, 'chat-guid')`
        '''
        
        return self.get_messages_interval(*args, **kwargs)


    @property
    def get_base_info(self) -> typing.Union[str, dict]:
        '''
        getting bases info from services base
        
        `self.get_base_info`
        '''

        return GetData.pwa(session=self.auth)


    def send_reaction(self, *args, **kwargs) -> None:
        '''
        this method has for send reaction on a message, coming soon for add in rubx
        plz wait for updating & adding method on rubika messenger
        '''

        pass


__all__ = ['UserMethods']


#!/bin/python
# rubika api bot methods

from .connection import GetData as http, Connection
import re, typing, logging


class BotMethods:


    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def __eq__(self, kwargs):
        '''
        from rb import RubikaClient
        from rb.responses.Bot import Text
        
        with RubikaClient(...) as client:
            print(`client == Text(chat_id=..., text=...))`
        '''
        
        if not isinstance(kwargs, dict):
            kwargs: dict = kwargs.dict()
            
        return self.send_message(**kwargs)
    
    def __ge__(self, kwargs):
        
        '''
        self >= dict(chat_id=..., text=...)
        '''

        if not isinstance(kwargs, dict):
            kwargs: dict = kwargs.dict()
        
        return self.edit_message_text(**kwargs)
    
    def __gt__(self, *args, **kwargs):
        
        '''
        self > 'chat-guid'
        '''
        
        return self.get_chat(*args)
    
    def __ne__(self, kwargs):
        
        '''
        self != dict(message_id=..., chat_id=...)
        '''
        
        if not isinstance(kwargs, dict):
            kwargs: dict = kwargs.dict()

        return self.delete_message(**kwargs)
    
    def __lt__(self, kwargs):
        
        '''
        self < dict(limit=1, offset_id=...)
        '''
        
        return self.get_updates(**kwargs)
    
    def __le__(self, kwargs):
        
        '''
        `self <= dict(...)`
        '''
        
        return self.edit_message_keypad(**kwargs)
    
    def __add__(self, kwargs):
        
        '''
        self + dict(from_chat_id=..., message_id=..., to_chat_id=...)
        '''

        if not isinstance(kwargs, dict):
            kwargs: dict = kwargs.dict()

        return self.forward_message(**kwargs)
    
    def __sub__(self, *args):
        
        '''
        self - 'file-id'
        '''
        
        return self.download(*args)
    
    def __mul__(self, kwargs):
        
        '''
        self * dict(...)
        '''
        
        return self.edit_chat_keypad(**kwargs)
    
    def __floordiv__(self, kwargs):
        
        '''
        self // dict(...)
        '''
        
        return self.send_files(**kwargs)
    
    def __truediv__(self, kwargs):
        
        '''
        self / dict(...)
        '''
        
        return self.send_poll(**kwargs)
    
    def __mod__(self, *args, **kwargs):
        
        '''
        `self % [str]`
        '''
        
        return self.set_commands(*args, **kwargs)
    
    def __pow__(self, kwargs):
        
        '''
        self ** dict(...)
        '''
        
        return self.send_sticker(**kwargs)

    def __lshift__(self, kwargs):
        
        '''
        self << dict(...)
        '''
        
        return self.send_location(**kwargs)
    
    def __rshift__(self, kwargs):
        '''
        self >> dict(chat_id=..., first_name=..., last_name=..., phone_number=...)
        '''
        return self.send_contact(**kwargs)
    
    def __and__(self, kwargs):
        
        '''
        self & dict(...)
        '''
        
        return self.update_bot_endpoint(**kwargs)
    
    def __or__(self, __type = 'me', /):
    
        '''
        `self | 'me'`
        '''
        
        return self.get_me

    def __xor__(self, kwargs):
    
        '''
        self ^ dict(...)
        '''
        
        return BotMethods.method(**kwargs)


    @classmethod
    def method(cls, token: str,
               method: str, **data) -> typing.Union[dict, str]:
        
        '''
        this method has for use from api bot to custom
        
        `self.method('token', 'sendMessage', chat_id=..., text=...)` # and other methods ...
        '''
        
        return http.bot(token=token, method=method, data=data)


    @property
    def get_me(self) -> typing.Union[dict, str]:
        '''
        getting your bot session
        
        `print(self.get_me)`
        '''
        
        return http.bot(token=self.token, method='getMe', data={})


    def get_chat(self, chat_id: str,
                 *args, **kwargs) -> typing.Union[str, dict]:
        
        '''
        get all message from a chat
        
        `self.get_chat('chat-id')`
        '''
        
        return http.bot(token=self.token, method='getChat', data={'chat_id': chat_id})


    def get_updates(self,
                    limit: int = 10,
                    offset_id: typing.Optional[str] = None,
                    *args,
                    **kwargs) -> typing.Union[str, dict]:
        
        '''
        `self.get_updates(1) # get last message`
        '''
        
        data: dict = {'limit': limit}.update({'offset_id': offset_id}) if offset_id else {'limit': limit}

        return http.bot(token=self.token, method='getUpdates', data=data)


    def update_bot_endpoint(
        self,
        url: str,
        types: typing.Literal['ReceiveUpdate', 'ReceiveInlineMessage',
                      'ReceiveQuery', 'GetSelectionItem',
                      'SearchSelectionItems']
        ) -> typing.Union[str, dict]:
        
        '''
        `self.update_bot_endpoint(...)`
        '''

        return http.bot(token=sslf.token, method='updateBotEndpoints', data={'url': url, 'type': types})


    def send_message(
        self,
        chat_id             : str,
        text                : str,
        chat_keypad         : str = None,
        disable_notification: bool = False,
        inline_keypad       : str = None,
        reply_to_message_id : str = None,
        chat_keypad_type    : typing.Literal['Remove', 'New', None] = None) -> typing.Union[str, dict]:
        
        '''
        this method for send message
        
        `self.send_message('chat-id', 'Hey!')`
        '''
        
        data: dict = {
            'text'                  :   text,
            'chat_id'               :   chat_id,
            'chat_keypad'           :   chat_keypad,
            'inline_keypad'         :   inline_keypad,
            'chat_keypad_type'      :   chat_keypad_type,
            'reply_to_message_id'   :   reply_to_message_id,
            'disable_notification'  :   disable_notification
            }

        return http.bot(token=self.token, method='sendMessage', data=data)


    def delete_message(self,
                       chat_id: str,
                       message_id: str) -> typing.Union[str, dict]:
        
        '''
        `self.delete_message('chat-id', '1234...') # delete a message from chat`
        '''
        
        return http.bot(token=self.token, method='delteMessage', data={'chat_id': chat_id, 'message_id': message_id})


    def forward_message(
        self,
        from_chat_id        : str,
        message_id          : str,
        to_chat_id          : str,
        disable_notification: bool = False
        ) -> typing.Union[str, dict]:
        
        '''
        `self.forward_message('f chat-id', '1234...', 't chat-id', True) # TODO: forward a message`
        '''
        
        data: dict = {
            'from_chat_id'          :   from_chat_id,
            'message_id'            :   message_id,
            'to_chat_id'            :   to_chat_id,
            'disable_notification'  :   disable_notification,
        }
        
        return http.bot(token=self.token, method='forwardMessage', data=data)


    def edit_message_text(self, chat_id: str,
                          message_id: str, text: str) -> typing.Union[str, dict]:
        
        '''
        `self.edit_message_text('chat-id', '1234...', 'new-event') # TODO: edit message for text type`
        '''
        
        data: dict = {
            'text'      :   text,
            'chat_id'   :   chat_id,
            'message_id':   message_id
            }

        return http.bot(token=self.token, method='editMessageText', data=data)


    def edit_message_keypad(self: 'BotMethods',
                            chat_id: str,
                            message_id: str,
                            inline_keypad: dict,
                            *args, **kwargs) -> typing.Union[str, dict]:
        
        '''
        `self.edit_message_keypad('chat-id', '1234...', {}) # TODO: edit message keypad`
        '''
        
        data: dict = {
            'chat_id'       :   chat_id,
            'message_id'    :   message_id,
            'inline_keypad' :   inline_keypad
            }

        return http.bot(token=self.token, method='editMessageKeypad', data=data)


    def edit_chat_keypad(
        self,
        chat_id         : str,
        chat_keypad     : dict = None,
        chat_keypad_type: typing.Literal['New', 'Remove'] = 'Remove',
        *args,
        **kwargs
        ) -> typing.Union[str, dict]:

        '''
        `self.edit_chat_keypad('chat-id'. chat_keypad_type='Remove') # TODO: remove`
        '''
        
        data: dict = {
            'chat_id'           :   chat_id,
            'chat_keypad_type'  :   chat_keypad_type,
            } if not chat_keypad else {'chat_id': chat_id, 'chat_keypad_type': chat_keypad_type, 'chat_keypad': chat_keypad}


        return http.bot(token=self.token, method='editChatKeypad', data=data)


    def send_poll(
        self,
        chat_id                 : str,
        question                : str,
        options                 : typing.List[str],
        chat_keypad             : dict = None,
        disable_notification    : bool = False,
        inline_keypad           : dict = None,
        reply_to_message_id     : typing.Optional[str] = None,
        chat_keypad_type        : typing.Literal[None, 'New', 'Remove'] = None
        ) -> typing.Union[dict, str]:
        
        '''
        this method for send poll (question)
        
        `self.send_poll('chat-id', 'how r u?', ['fine', 'bad'])`
        '''
        
        data: dict = {
            'chat_id'               :   chat_id,
            'options'               :   options,
            'question'              :   question,
            'chat_keypad'           :   chat_keypad,
            'inline_keypad'         :   inline_keypad,
            'chat_keypad_type'      :   chat_keypad_type,
            'reply_to_message_id'   :   reply_to_message_id,
            'disable_notification'  :   disable_notification
            }
        
        return http.bot(token=self.token, method='sendPoll', data=data)


    def send_location(
        self,
        token               : str,
        chat_id             : str,
        latitude            : str,
        longitude           : str,
        chat_keypad         : dict = None,
        disable_notification: bool = False,
        inline_keypad       : dict = None,
        reply_to_message_id : typing.Optional[str] = None,
        chat_keypad_type    : typing.Literal[None, 'New', 'Remove'] = None,
        ) -> typing.Union[str, dict]:
        
        '''
        this method has for send location
        
        `self.send_location('chat-id', ...)`
        '''
        
        data: dict = {
            'chat_id'               :   chat_id,
            'latitude'              :   latitude,
            'longitude'             :   longitude,
            'chat_keypad'           :   chat_keypad,
            'disable_notification'  :   disable_notification,
            'inline_keypad'         :   inline_keypad,
            'reply_to_message_id'   :   reply_to_message_id,
            'chat_keypad_type'      :   chat_keypad_type
            }
        
        return http.bot(token=self.token, method='sendLocation', data=data)


    def send_sticker(
        self,
        chat_id             : str,
        sticker_id          : str,
        chat_keypad         : dict = None,
        disable_notification: bool = False,
        inline_keypad       : dict = None,
        reply_to_message_id : typing.Optional[str] = None,
        chat_keypad_type    : typing.Literal[None, 'New', 'Remove'] = None
        ) -> typing.Union[str, dict]:
        
        '''
        sent a sticker to chat
        
        `self.send_sticker('chat-id', 's-id')`
        '''
        
        data: dict = {
            'chat_id'               :   chat_id,
            'sticker_id'            :   sticker_id,
            'chat_keypad'           :   chat_keypad,
            'disable_notification'  :   disable_notification,
            'inline_keypad'         :   inline_keypad,
            'reply_to_message_id'   :   reply_to_message_id,
            'chat_keypad_type'      :   chat_keypad_type
            }
        
        return http.bot(token=self.token, method='sendSticker', data=data)


    def send_contact(
        self,
        chat_id             : str,
        first_name          : str,
        last_name           : str,
        phone_number        : str,
        chat_keypad         : dict = None,
        disable_notification: bool = False,
        inline_keypad       : dict = None,
        reply_to_message_id : typing.Optional[str] = None,
        chat_keypad_type    : typing.Literal[None, 'New', 'Remove'] = None
        ) -> typing.Union[str, dict]:
        
        data: dict = {
            'chat_id'               :   chat_id,
            'last_name'             :   last_name,
            'first_name'            :   first_name,
            'chat_keypad'           :   chat_keypad,
            'phone_number'          :   phone_number,
            'inline_keypad'         :   inline_keypad,
            'chat_keypad_type'      :   chat_keypad_type,
            'reply_to_message_id'   :   reply_to_message_id,
            'disable_notification'  :   disable_notification
            }
        
        return http.bot(token=self.token, method='sendContact', data=data)


    def __request_send_file(
        self,
        types: typing.Literal['File', 'Image',
                               'Voice', 'Music', 'Gif']
        ) -> typing.Union[str, dict]:
        
        '''
        get url upload file from api
        
        `self.__request_send_file('Music') # return a dict or attr for get url`
        '''
        
        return http.bot(token=self.token, method='requestSendFile', data={'type': types})


    def send_files(
        self,
        chat_id             :   str,
        file                :   str                                                      = None,
        file_type           :   typing.Literal['File', 'Image', 'Voice', 'Music', 'Gif'] = 'File',
        chat_keypad         :   dict                                                     = None,
        disable_notification:   bool                                                     = False,
        inline_keypad       :   dict                                                     = None,
        reply_to_message_id :   typing.Optional[str]                                     = None,
        chat_keypad_type    :   typing.Literal[None, 'New', 'Remove']                    = None
        ) -> typing.Union[str, dict]:


        '''
        # this method has for sending all format type file
        ## make uploading
        
        chat_id: object id
        
        file: file path
        
        file_type: for file action type = 'File', 'Image', 'Voice', 'Music', 'Gif'
        
        chat_keypad: {'rows': {'buttons': []}, 'resize_keyboard': False, 'on_time_keyboard': True}
        
        disable_notification: for see user notif: False, True
        
        inline_keypad: chat_keypad inline
        
        reply_to_message_id: for reply to a message
        
        chat_keypad_type: if u usage chat_keypad: insert action type: 'New', 'Remove'


        # Use:
            - `self.send_files('chat-guid', 'C:\\music.mp3', 'Music')` 
        '''

        url = self.__request_send_file(file_type or 'File').get('download_url')
        file_id = self.__upload(url, file).get('file_id')

        data: dict = {
            'chat_id'               :   chat_id,
            'file_id'               :   file_id,
            'chat_keypad'           :   chat_keypad,
            'disable_notification'  :   disable_notification,
            'inline_keypad'         :   inline_keypad,
            'reply_to_message_id'   :   reply_to_message_id,
            'chat_keypad_type'      :   chat_keypad_type
            }
        
        return http.bot(token=self.token, method='sendFile', data=data)


    def __upload(self, url: str, file_name: str,
               file_path: str = None, *args, **kwargs) -> typing.Union[str, dict]:
        
        '''
        get file id from url
        
        `self.__upload('url', 'C:/file.mp3')`
        '''
        
        return http.bot('file', token=self.token, data={'file': (file_name.split('\\')[-1] if file_name.__contains__('\\') else file_name.split('/')[-1] if file_name.__contains__('/') else file_name, open(file_path or file_name, 'rb+'))}, url=url)


    def download(self, file_id: str) -> typing.Union[str, dict]:
        '''
        downloading file from id
        
        `self.download('file-id')`
        '''
        
        return http.bot(token=self.token, method='getFile', data={'file_id': file_id})


    def get_file(self, file_id: str,
                 *args, **kwargs) -> typing.Union[str, dict]:
        
        '''
        downloading file from id
        
        `self.get_file('file-id')`
        '''
        
        return self.download(file_id)


    def set_commands(self, bot_commands: typing.List[str],
                     *args, **kwargs) -> typing.Union[str, dict]:
        
        '''
        `self.set_commands([str]) # TODO: set bot commands`
        '''
        
        return http.bot(token=self.token, method='setCommands', data={'bot_commands': bot_commands})


    def on(self) -> None:
        pass

__all__ = ['BotMethods']


#!/bin/python

from Crypto.Cipher.AES      import MODE_CBC, block_size, new
from Crypto.Util.Padding    import pad, unpad
from base64                 import b64encode, urlsafe_b64decode

class Encryption(object):
    
    @classmethod
    
    def __init__(cls, auth_key: str) -> ...:
        if len(str(auth_key)) != 32:
            raise IndexError('len your session key is not \'32\'')
        cls.auth_key, cls.iv = bytearray((''.join(list(map(lambda a: chr((a - ord('a') + 9) % 26 + ord('a')), (str(auth_key[16:24] + auth_key[:8] + auth_key[24:32] + auth_key[8:16]).encode('latin-1')))))), 'utf-8'), bytearray.fromhex(str('0' * len(auth_key)))

    def encrypt(cls, data: str) -> (str):
        return (b64encode(new(cls.auth_key, MODE_CBC, iv=cls.iv).encrypt(pad(data.encode('utf-8'), block_size))).decode('utf-8'))

    def decrypt(cls, data: str) -> (str):
        return (unpad(new(cls.auth_key, MODE_CBC, iv=cls.iv).decrypt(urlsafe_b64decode(data.encode('utf-8'))), block_size)).decode('utf-8')




from .tools import (UpToDate, _Top, Clean,
                    Scanner, Maker, Login, Thumbnail)
from .installer import PyPi

__all__ = ['UpToDate', '_Top', 'Clean',
           'Scanner', 'Maker', 'Login',
           'Thumbnail', 'PyPi']


#!/bin/python


class Device(dict):

    '''
    this is a default device for account register
    '''

    DEFAULT_DEVICE: dict = {
        'app_version'           :   'MA_2.9.8',
        'device_hash'           :   'CEF34215E3E610825DC1C4BF9864D47A',
        'device_model'          :   'rubx-lib',
        'is_multi_account'      :   False,
        'lang_code'             :   'fa',
        'system_version'        :   'SDK 22',
        'token'                 :   'cgpzI3mbTPKddhgKQV9lwS:APA91bE3ZrCdFosZAm5qUaG29xJhCjzw37wE4CdzAwZTawnHZM_hwZYbPPmBedllAHlm60v5N2ms-0OIqJuFd5dWRAqac2Ov-gBzyjMx5FEBJ_7nbBv5z6hl4_XiJ3wRMcVtxCVM9TA-',
        'token_type'            :   'Firebase'
    }


class Infos:

    '''
    info extends
    '''

    citys, proxys, auth_, sent = [], [], [], lambda data: False if data.get('status').lower() != 'ok' else True


class clients(dict):

    '''
    rubika client platforms
    '''

    (web, android,
     rubx, pwa) = (
         {
                'app_name'      :   'Main',
                'app_version'   :   '4.2.0',
                'platform'      :   'Web',
                'package'       :   'web.rubika.ir',
                'lang_code'     :   'fa'
                },
            {
                'app_name'      :   'Main',
                'app_version'   :   '2.8.1',
                'platform'      :   'Android',
                'package'       :   'ir.resaneh1.iptv',
                'lang_code'     :   'fa'
                },
            {
                'app_name'      :   'Main',
                'app_version'   :   '3.0.8',
                'platform'      :   'Android',
                'package'       :   'app.rbmain.a',
                'lang_code'     :   'en'
                },
            {
                'app_name'      :   'Main',
                'app_version'   :   '4.2.0',
                'platform'      :   'PWA',
                'package'       :   'web.rubika.ir'
                }
        )


__all__ = ['Device', 'Infos', 'clients']


#!/bin/python

from setuptools import find_packages, setup

requires = ['requests', 'urllib3', 'datetime']
version = '10.5.5'

readme = '''
<p align="center">
    <a href="https://github.com/mester-root/rubx">
        <img src="https://raw.githubusercontent.com/Mester-Root/rubx/main/icons/rubx-action.png" alt="RuBx" width="420">
    </a>
    <br>
    <b>Rubika Client API Framework | Python 3</b>
    <br>
    <a href="https://github.com/Mester-Root/rubx/blob/main/README.md">
        Document
    </a>
     •
    <a href="https://t.me/rubx_library">
        Telegram
    </a>
     •
    <a href="https://rubika.ir/TheClient">
        Rubika
    </a>
</p>


# [Rubx] - rubika messenger

## Messenger:

```python

from rb import RubikaClient # rb: is main package

with RubikaClient('session') as client:
   client.send_message('Hello From Rubx', 'chat-guid')

```

### Or

```python
from rb import RubikaClient as Client

def respond(callable, params) -> dict:
    return callable(**params)

with Client(...) as client:
    print(
        respond(
            client.send_message,
            dict(
                chat_id=...,
                text='**Hey** @User@ R U __My__ ``Friend`` and ~~My Love~~'
                mention_user_ids=['u0...']),
            )
        )
    
    # print(respond(client.send_message, dict(chat_id=..., sticker=True, emoji_character='😜', sticker_id=..., )))
```

### Or

## shorcuts

```python
from rb import RubikaClient
from rb.responses import Self

with RubikaClient(...) as client:
    print(client == Self.Text(text='Hey', chat_id='chat-guid')) # to send message
    # print(client * 'chat-guid') # to get chat info
    # use the operators

```

### for: if you forget the method name
```python
from rb import RubikaClient

with RubikaClient('session') as client:
    print(client.getChatInfo(client, 'chat-guid')) # GetChatInfo, GETchatINFO, or ...
    # normally: client.get_chat_info('chat-guid')
```

#### for show all operators and shortcuts go to docs


## Rubino

``` python

from rb import RubinoClient

with RubinoClient('session') as app:
    app.create_page(...)

```

## Handler
### Handler Examples

```python
from rb import Handler, EventBuilder, Filters

client = Handler(...)

# handlers: HandShake, ChatsUpdates, MessagesUpdates
client.add_event_handling(func='ChatsUpdates', events=dict(get_chats=True, get_messages=True, pattern=('/start', 'Hey from rubx lib.')))

@client.handler
def hello(app, message: EventBuilder, event):
    # to print message: print(message) or print(event)
    # to use all methods: app.create_objcet_voice_chat(...)
    message.respond(message.pattern, Filters.author) # filters: chat, group, channel, author
```

### Or

```python
from rb import Handler, NewMessage, Filters, EventBuilder, Performers

client = Handler(...)

@client.on(NewMessage(client.handle, handle_name=Performers.chats_updates).builder())
def update(event: EventBuilder):
    ... # event.respond('Hey', Filters.chat)

```

## Or

```python
from rb import Handler, Filters, Performers

client = Handler('session')

def event(message):
    message.respond(message.pattern, Filters.author)
        
client.add_event_handling(func=Performers.hand_shake, events=dict(get_chats=True, get_messages=True, pattern=('/start', 'Hi from rubx lib.')))
client.starting = True
client.command_handler(event)
```

## To using HandShake(WebSocket):

```python
from rb import Handler, EventBuilder, Filters, Performers

client = Handler('abc...', 'u0...')
client.add_event_handling(func=Performers.hand_shake, events=dict(get_messages=True, get_chats=False))
@client.handler
def update(app, update, event):
    if update.message.text == '/start':
        message.reply(text='Hello my dear', chat_id=update.message.author_object_guid, reply_to_message_id=update.message.message_id)
        # or using repond: message.respond('Hey!', Filters.author)
```

## Async methods

```python
from rb import Client # Client: asycn reader

async def run(*args):
    async with Client(...) as client:
        result = await client.start(client.send_message, 'Hey! from rubx', 'chat-guid')
        print(result)

Client.run(run)
```

### How to get keys from result?

```python
from rb import RubikaClient

with RubikaClient('session', return_data_action='str') as client:
    print((client * 'chat-guid').data.chat) # get chat info: get data and get chat from a chat object
```


## Bot API Methods

### example for api methods send message text
```python
from rb import BotAPI

with BotAPI(__name__, 'token') as app:
    app.send_message('chat-id', 'Hey!')
```

### Handler
```python
from rb import BotAPI

with BotAPI(__name__, 'token') as app:
    app.add_event_handling(('\w{1}start', 'Hello'))
    
    @app.handler
    def update(methods, update, event):
        ...
```


_____________________________

Rubx - ⚡
========

  - Now the best ‍`sync‍` and `asycn` library for Rubika's was developed
  - ⭐️ Thanks **everyone** who has starred the project, it means a lot!

**Rubx** is an sync **Python 3** rubika library to interact with Rubika's API
as a user or through a bot account (self API alternative).

    🔴 If you have code using Rubx before its 8.0.5 version, you must
    read docs to learn how to migrate. 💡

What is this?
-------------

🇮🇷 - Rubika is a popular messaging application. This library is meant
to make it easy for you to write Python programs that can interact
with Rubika. Think of it as a wrapper that has already done the
heavy job for you, so you can focus on developing an application.
This module provides all the desired methods with a very simple and beautiful user interface and has a very high speed.
Give your employer the best experience of a project.


Updating - 🌀 :
--------
    - Complete documentation and optimization.

___________________________


## INSTALLING
```bash
pip install rubx
```


## UPGRADE
```
pip install rubx --upgrade
```

## CREATED BY:

    - saleh


# self rubika client with python3 RUBX module ![](https://i.imgur.com/fe85aVR.png)

_______________________

[![Python 3|2.7|3.x](https://img.shields.io/badge/python-3|3.0|3.x-yellow.svg)](https://www.python.org/)

[![License](https://img.shields.io/badge/license-GPLv2-red.svg)](https://raw.githubusercontent.com/Mester-Root/rubx/main/LICENSE)

[![Creator](https://img.shields.io/badge/Telegram-Channel-33A8E3)](https://t.me/rubx_library)

[![Telegram](https://img.shields.io/badge/-telegram-red?color=white&logo=telegram&logoColor=black)](https://t.me/clientUser)
_______________________

<div align="center">

![issues](https://img.shields.io/github/issues/mester-root/rubx)
![forks](https://img.shields.io/github/forks/mester-root/rubx)
![version](https://img.shields.io/badge/version-v--1.0.1--beta-yellow)
![stars](https://img.shields.io/github/stars/mester-root/rubx)
![license](https://img.shields.io/github/license/mester-root/rubx)
![icon](https://raw.githubusercontent.com/Mester-Root/rubx/main/logo.png)
</div>


_______________________


### **special**:

- *[RUBX] > a library 'official' for rubika messnger with client server.*
- *[RUBX] > use api's rubika, and full methods.*

#### to use it, your python must support **type hint**.
'''

setup(
    name="rubx",
    version=version,
    description="Rubx Library For Rubika Iranina Messenger",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/mester-root/rubx",
    download_url="https://github.com/mester-root/rubx/releases/latest",
    author="Saleh",
    author_email="m3st3r.r00t@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet",
        "Topic :: Communications",
        "Topic :: Communications :: Chat",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks"
    ],
    keywords=['messenger', 'python', 'self', 'rubx', 'rubix', 'rubikax', 'rubika', 'bot', 'robot', 'library', 'rubikalib', 'rubikalibrary', 'rubika.ir', 'web.rubika.ir', 'telegram'],
    project_urls={
        "Tracker": "https://github.com/mester-root/rubx/issues",
        "Channel": "https://t.me/rubx_library",
        "Source": "https://github.com/mester-root/rubx",
        "Documentation": "https://github.com/Mester-Root/rubx/blob/main/README.md",
    },
    python_requires="~=3.9",
    packages=find_packages(),
    zip_safe=False,
    install_requires=requires
)


#!/bin/python

import struct, typing, os
from datetime       import datetime, date, timedelta, timezone
from re             import compile, escape, search, findall, sub
from time           import sleep, gmtime, localtime
from base64         import b64encode
from collections    import deque
from html.parser    import HTMLParser


'''
parse message text to markdown and html
'''


_EPOCH_NAIVE = datetime(*gmtime(0)[:6])
_EPOCH_NAIVE_LOCAL = datetime(*localtime(0)[:6])
_EPOCH = _EPOCH_NAIVE.replace(tzinfo=timezone.utc)


def _datetime_to_timestamp(dt):
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    secs = int((dt - _EPOCH).total_seconds())
    return struct.unpack('i', struct.pack('I', secs & 0xffffffff))[0]


def _json_default(value):
    if isinstance(value, bytes):
        return b64encode(value).decode('ascii')
    elif isinstance(value, datetime):
        return value.isoformat()
    else:
        return repr(value)


class TLObject:
    CONSTRUCTOR_ID = None
    SUBCLASS_OF_ID = None

    @staticmethod
    def pretty_format(obj, indent=None):
        if indent is None:
            if isinstance(obj, TLObject):
                obj = obj.to_dict()

            if isinstance(obj, dict):
                return '{}({})'.format(obj.get('_', 'dict'), ', '.join(
                    '{}={}'.format(k, TLObject.pretty_format(v))
                    for k, v in obj.items() if k != '_'
                ))
            elif isinstance(obj, str) or isinstance(obj, bytes):
                return repr(obj)
            elif hasattr(obj, '__iter__'):
                return '[{}]'.format(
                    ', '.join(TLObject.pretty_format(x) for x in obj)
                )
            else:
                return repr(obj)
        else:
            result = []
            if isinstance(obj, TLObject):
                obj = obj.to_dict()

            if isinstance(obj, dict):
                result.append(obj.get('_', 'dict'))
                result.append('(')
                if obj:
                    result.append('\n')
                    indent += 1
                    for k, v in obj.items():
                        if k == '_':
                            continue
                        result.append('\t' * indent)
                        result.append(k)
                        result.append('=')
                        result.append(TLObject.pretty_format(v, indent))
                        result.append(',\n')
                    result.pop()  # last ',\n'
                    indent -= 1
                    result.append('\n')
                    result.append('\t' * indent)
                result.append(')')

            elif isinstance(obj, str) or isinstance(obj, bytes):
                result.append(repr(obj))

            elif hasattr(obj, '__iter__'):
                result.append('[\n')
                indent += 1
                for x in obj:
                    result.append('\t' * indent)
                    result.append(TLObject.pretty_format(x, indent))
                    result.append(',\n')
                indent -= 1
                result.append('\t' * indent)
                result.append(']')

            else:
                result.append(repr(obj))

            return ''.join(result)

    @staticmethod
    def serialize_bytes(data):
        if not isinstance(data, bytes):
            if isinstance(data, str):
                data = data.encode('utf-8')
            else:
                raise TypeError(
                    'bytes or str expected, not {}'.format(type(data)))

        r = []
        if len(data) < 254:
            padding = (len(data) + 1) % 4
            if padding != 0:
                padding = 4 - padding

            r.append(bytes([len(data)]))
            r.append(data)

        else:
            padding = len(data) % 4
            if padding != 0:
                padding = 4 - padding

            r.append(bytes([
                254,
                len(data) % 256,
                (len(data) >> 8) % 256,
                (len(data) >> 16) % 256
            ]))
            r.append(data)

        r.append(bytes(padding))
        return b''.join(r)

    @staticmethod
    def serialize_datetime(dt):
        if not dt and not isinstance(dt, timedelta):
            return b'\0\0\0\0'

        if isinstance(dt, datetime):
            dt = _datetime_to_timestamp(dt)
        elif isinstance(dt, date):
            dt = _datetime_to_timestamp(datetime(dt.year, dt.month, dt.day))
        elif isinstance(dt, float):
            dt = int(dt)
        elif isinstance(dt, timedelta):
            dt = _datetime_to_timestamp(datetime.utcnow() + dt)

        if isinstance(dt, int):
            return struct.pack('<i', dt)

        raise TypeError('Cannot interpret "{}" as a date.'.format(dt))

    def __eq__(self, o):
        return isinstance(o, type(self)) and self.to_dict() == o.to_dict()

    def __ne__(self, o):
        return not isinstance(o, type(self)) or self.to_dict() != o.to_dict()

    def __str__(self):
        return TLObject.pretty_format(self)

    def stringify(self):
        return TLObject.pretty_format(self, indent=0)

    def to_dict(self):
        raise NotImplementedError

    def to_json(self, fp=None, default=_json_default, **kwargs):
        d = self.to_dict()
        if fp:
            return json.dump(d, fp, default=default, **kwargs)
        else:
            return json.dumps(d, default=default, **kwargs)

    def __bytes__(self):
        try:
            return self._bytes()
        except AttributeError:
            raise TypeError('a TLObject was expected but found something else')
    def _bytes(self):
        raise NotImplementedError

    @classmethod
    def from_reader(cls, reader):
        raise NotImplementedError


class MessageEmpty(TLObject):
    CONSTRUCTOR_ID = 0x90a6ca84
    SUBCLASS_OF_ID = 0x790009e3

    def __init__(self, id: int, peer_id: typing.Optional['TypePeer']=None):
        self.id = id
        self.peer_id = peer_id

    def to_dict(self):
        return {
            '_': 'MessageEmpty',
            'id': self.id,
            'peer_id': self.peer_id.to_dict() if isinstance(self.peer_id, TLObject) else self.peer_id
        }

    def _bytes(self):
        return b''.join((
            b'\x84\xca\xa6\x90',
            struct.pack('<I', (0 if self.peer_id is None or self.peer_id is False else 1)),
            struct.pack('<i', self.id),
            b'' if self.peer_id is None or self.peer_id is False else (self.peer_id._bytes()),
        ))

    @classmethod
    def from_reader(cls, reader):
        flags = reader.read_int()

        _id = reader.read_int()
        if flags & 1:
            _peer_id = reader.tgread_object()
        else:
            _peer_id = None
        return cls(id=_id, peer_id=_peer_id)


class MessageEntityBankCard(TLObject):
    CONSTRUCTOR_ID = 0x761e6af4
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityBankCard',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'\xf4j\x1ev',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityBold(TLObject):
    CONSTRUCTOR_ID = 0xbd610bc9
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityBold',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'\xc9\x0ba\xbd',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityCode(TLObject):
    CONSTRUCTOR_ID = 0x28a20571
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityCode',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'q\x05\xa2(',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityHashtag(TLObject):
    CONSTRUCTOR_ID = 0x6f635b0d
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityHashtag',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'\r[co',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityItalic(TLObject):
    CONSTRUCTOR_ID = 0x826f8b60
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityItalic',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'`\x8bo\x82',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityMention(TLObject):
    CONSTRUCTOR_ID = 0xfa04579d
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityMention',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'\x9dW\x04\xfa',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityMentionName(TLObject):
    CONSTRUCTOR_ID = 0xdc7b1140
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int, user_id):
        self.offset = offset
        self.length = length
        self.user_id = user_id

    def to_dict(self):
        return {
            '_': 'MessageEntityMentionName',
            'offset': self.offset,
            'length': self.length,
            'user_id': self.user_id
        }

    def _bytes(self):
        return b''.join((
            b'@\x11{\xdc',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
            struct.pack('<q', self.user_id),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        _user_id = reader.read_long()
        return cls(offset=_offset, length=_length, user_id=_user_id)


class MessageEntityPre(TLObject):
    CONSTRUCTOR_ID = 0x73924be0
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int, language: str):
        self.offset = offset
        self.length = length
        self.language = language

    def to_dict(self):
        return {
            '_': 'MessageEntityPre',
            'offset': self.offset,
            'length': self.length,
            'language': self.language
        }

    def _bytes(self):
        return b''.join((
            b'\xe0K\x92s',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
            self.serialize_bytes(self.language),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        _language = reader.tgread_string()
        return cls(offset=_offset, length=_length, language=_language)


class MessageEntityStrike(TLObject):
    CONSTRUCTOR_ID = 0xbf0693d4
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityStrike',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'\xd4\x93\x06\xbf',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityTextUrl(TLObject):
    CONSTRUCTOR_ID = 0x76a6d327
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int, url: str):
        self.offset = offset
        self.length = length
        self.url = url

    def to_dict(self):
        return {
            '_': 'MessageEntityTextUrl',
            'offset': self.offset,
            'length': self.length,
            'url': self.url
        }

    def _bytes(self):
        return b''.join((
            b"'\xd3\xa6v",
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
            self.serialize_bytes(self.url),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        _url = reader.tgread_string()
        return cls(offset=_offset, length=_length, url=_url)


class MessageEntityUnderline(TLObject):
    CONSTRUCTOR_ID = 0x9c4e7e8b
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityUnderline',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'\x8b~N\x9c',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityUnknown(TLObject):
    CONSTRUCTOR_ID = 0xbb92ba95
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityUnknown',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'\x95\xba\x92\xbb',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)


class MessageEntityUrl(TLObject):
    CONSTRUCTOR_ID = 0x6ed02538
    SUBCLASS_OF_ID = 0xcf6419dc

    def __init__(self, offset: int, length: int):
        self.offset = offset
        self.length = length

    def to_dict(self):
        return {
            '_': 'MessageEntityUrl',
            'offset': self.offset,
            'length': self.length
        }

    def _bytes(self):
        return b''.join((
            b'8%\xd0n',
            struct.pack('<i', self.offset),
            struct.pack('<i', self.length),
        ))

    @classmethod
    def from_reader(cls, reader):
        _offset = reader.read_int()
        _length = reader.read_int()
        return cls(offset=_offset, length=_length)

    def _bytes(self):
        return b''.join((
            b'[\x1bf\xab',
        ))

    @classmethod
    def from_reader(cls, reader):
        return cls()


DEFAULT_URL_RE = compile(r'\[([\S\s]+?)\]\((.+?)\)')
DEFAULT_URL_FORMAT = '[{0}]({1})'
DEFAULT_DELIMITERS = {
    '**': MessageEntityBold,
    '__': MessageEntityItalic,
    '~~': MessageEntityStrike,
    '``': MessageEntityCode,
    '```': MessageEntityPre,
    '@': MessageEntityMention,
    '#': MessageEntityHashtag,
    '_': MessageEntityUnderline
    }


class MarkDown(object):

    def __init__(self, *args) -> (None):
        
        '''
        text cleaner to markdown
        '''
        
        pass

    def add_surrogate(self, text):
        return ''.join(
            ''.join(chr(y) for y in struct.unpack('<HH', x.encode('utf-16le')))
            if (0x10000 <= ord(x) <= 0x10FFFF) else x for x in text
        )

    def del_surrogate(self, text):
        return text.encode('utf-16', 'surrogatepass').decode('utf-16')

    def strip_text(self, text, entities):
        if not entities:
            return text.strip()

        while text and text[-1].isspace():
            e = entities[-1]
            if e.offset + e.length == len(text):
                if e.length == 1:
                    del entities[-1]
                    if not entities:
                        return text.strip()
                else:
                    e.length -= 1
            text = text[:-1]

        while text and text[0].isspace():
            for i in reversed(range(len(entities))):
                e = entities[i]
                if e.offset != 0:
                    e.offset -= 1
                    continue

                if e.length == 1:
                    del entities[0]
                    if not entities:
                        return text.lstrip()
                else:
                    e.length -= 1

            text = text[1:]

        return text

    def parse(self, message, delimiters=None, url_re=None):
        if not message:
            return message, []

        if url_re is None:
            url_re = DEFAULT_URL_RE
        elif isinstance(url_re, str):
            url_re = compile(url_re)

        if not delimiters:
            if delimiters is not None:
                return message, []
            delimiters = DEFAULT_DELIMITERS

        delim_re = compile('|'.join('({})'.format(escape(k))
                                for k in sorted(delimiters, key=len, reverse=True)))
        i = 0
        result = []
        message = self.add_surrogate(message)
        while i < len(message):
            m = delim_re.match(message, pos=i)
            if m:
                delim = next(filter(None, m.groups()))
                end = message.find(delim, i + len(delim) + 1)
                if end != -1:
                    message = ''.join((
                            message[:i],
                            message[i + len(delim):end],
                            message[end + len(delim):]
                    ))
                    for ent in result:
                        if ent.offset + ent.length > i:
                            if ent.offset <= i:
                                ent.length -= len(delim) * 2
                            else:
                                ent.length -= len(delim)
                    ent = delimiters[delim]
                    if ent == MessageEntityPre:
                        result.append(ent(i, end - i - len(delim), ''))  # has 'lang'
                    else:
                        result.append(ent(i, end - i - len(delim)))

                    if ent in (MessageEntityCode, MessageEntityPre):
                        i = end - len(delim)

                    continue

            elif url_re:
                m = url_re.match(message, pos=i)
                if m:
                    message = ''.join((
                        message[:m.start()],
                        m.group(1),
                        message[m.end():]
                    ))

                    delim_size = m.end() - m.start() - len(m.group())
                    for ent in result:
                        if ent.offset + ent.length > m.start():
                            ent.length -= delim_size

                    result.append(MessageEntityTextUrl(
                        offset=m.start(), length=len(m.group(1)),
                        url=del_surrogate(m.group(2))
                    ))
                    i += len(m.group(1))
                    continue

            i += 1

        message = self.strip_text(message, result)
        return self.del_surrogate(message), result


class Metas(object):
    
    def __init__(self, text: str, user_ids,
                 action: bool = True, *args, **kwargs) -> (None):
        
        
        '''
        to metadata parser - bold, italic, mono, strike and mention for message text
        
        return data an tuple to main text dont enity and client data
        '''
        
        self.text, self.user_ids, self.action = text, user_ids, action

    def __replacer(self, text: str, action: bool = 'regex', mode: bool = True) -> (str):

        if action == 'regex':
            for word in (compile(r'\_\_'), compile(r'\*\*'), compile(r'\`\`')):
                text: str = sub(word, '', text)
            else:
                return text

        elif action == 'replace':
            
            #text: str = text.replace('____', '')
            plorer, index_list = search(r'(\@.*?\_\_\w+\_\_\w+)|(\@.*?\_\_\_\_\w+)', text), []

            if mode:
                text: list = [*text]
                for i, n in zip(text, range(len(text))):
                    try:
                        if i == '_' and text[n+1 if not len(text) == n + 1 else n-1] != '_' and text[n-1 if n != 0 else n+1] != '_':
                            index_list.append(n)
                            text[n] = ''
                    except IndexError:
                        pass
                else:
                    text: str = ''.join(text)

            if plorer:

                finds, main_index_list, texts = findall(r'(\@.*?\_\_\w+\_\_\w+)|(\@.*?\_\_\_\_\w+)', text), [], text

                for find in finds:
                    if isinstance(find, tuple) or isinstance(find, list):
                        find: str = ''.join(list(find))
                        if find == '':
                            break
                    for word in ('__', '``', '**'):
                        if [*texts].count(word[0]) >= 4:
                            if str([*texts].count(word[0]) / 4).split('.')[1] == '0':
                                texts: str = texts.replace(word, '')
                            else:
                                texts: str = texts.replace(word, '', int([*texts].count(word[0]) / 2) - 1)
                    else:
                        main_index_list.append(texts.find(find.replace('__', '')))
                    text: str = text.replace(find, '', 1)

            for word in ('__', '**', '``'):
                if [*text].count(word[0]) >= 4:
                    if str([*text].count(word[0]) / 4).split('.')[1] == '0':
                        text: str = text.replace(word, '')
                    else:
                        text: str = text.replace(word, '', int([*text].count(word[0]) / 2) - 1)
            else:

                if plorer:

                    texts: list = [*text]

                    for index, find, num in zip(main_index_list, finds, range(len(finds))):
                        if isinstance(find, tuple) or isinstance(find, list):
                            find: str = ''.join(list(find))
                        texts.insert(index, find)
                    if mode:
                        list(map(lambda number: texts.insert(number, '_'), index_list)) # int(len(findall(r'\_', text)) ** 2)
                    
                    return ''.join(texts)
                else:
                    return text

        else:
            return text.replace('**', '', len(findall(r'\*\*(.*?)\*\*', text)) + 1).replace('__', '', len(findall(r'(\_\_(.*?)\_\_)', text)) + 1).replace('``', '', len(findall(r'\`\`(.*?)\`\`', text)) + 1)


    def replacer(self, text: str, action: str = 'replace'):
        return self.__replacer(text, action)


    def __to_parse(self, text: str, user_ids: list = None, filter_hashtag: bool = True, filter_under_line: bool = True) -> (tuple):

        if not filter_hashtag:
            DEFAULT_DELIMITERS.pop('#')
        if not filter_under_line:
            DEFAULT_DELIMITERS.pop('_')

        mark, params, action, count = MarkDown(), [], '', 0

        for text in (mark.parse(text)):

            if isinstance(text, str):
                real_text: str = text

            else:

                for clean in text:

                    if isinstance(clean, MessageEntityBold):
                        action: str = 'Bold'
                    elif isinstance(clean, MessageEntityItalic):
                        action: str = 'Italic'
                    elif isinstance(clean, MessageEntityStrike):
                        action: str = 'Strike'
                    elif isinstance(clean, MessageEntityCode):
                        action: str = 'Mono'
                    elif  isinstance(clean, MessageEntityPre):
                        action: str = 'Pre'
                    elif isinstance(clean, MessageEntityMention):
                        action: str = 'MentionText'
                    
                    if not isinstance(clean, MessageEntityMention):
                        params.extend([{'from_index': clean.offset, 'length': clean.length, 'type': action}])
                    
                    else:
                        guid: str = user_ids[count]
                        params.extend([{'type': action, 'mention_text_object_guid': guid, 'from_index': clean.offset, 'length': clean.length, 'mention_text_object_type': 'User'}])
                        count: int = count.__add__(1)
                
                else:
                    return params, real_text


    @property
    def checker(self) -> (list):
        
        if self.user_ids or self.action:
            return self.__to_parse(self.text, self.user_ids, False, False)
        else:
            result, texts = [], self.__replacer(self.text, action='replace')

            if ('**' in self.text):
                
                Bold: list = findall(r'\*\*(.*?)\*\*', self.text)
                boldFromIndex: list = [self.text.index(i) for i in Bold]
                [(result.append({'from_index': from_index, 'length': len(length), 'type': 'Bold'})) for from_index, length in zip(boldFromIndex, Bold)]
            
            if ('__' in self.text):
                
                Italic: list = findall(r'\_\_(.*?)\_\_', self.text)
                ItalicFromIndex: list = [self.text.index(i) for i in Italic]
                [(result.append({'from_index': from_index, 'length': len(length), 'type': 'Italic'})) for from_index, length in zip(ItalicFromIndex, Italic)]
            
            if ('``' in self.text):
                
                Mono: list = findall(r'\`\`(.*?)\`\`', self.text)
                monoFromIndex: list = [self.text.index(i) for i in Mono]
                [self.text.index(i) for i in Mono]
                [(result.append({'from_index': from_index, 'length': len(length), 'type': 'Mono'})) for from_index, length in zip(monoFromIndex, Mono)]
            
            return result, texts


class Tags(object):
    
    def __init__(self, text: str, *args) -> (dict):
        
        '''
        for mention [tag] text message with `@`
    
        '''
        
        self.text: str = text

    def __replacer(self, text: str) -> (str):
        return sub(compile(r'\@'), '', text)

    def checker(
        self,
        guids: list =   None,
        types: list =   None
        ) -> (tuple):
        
        if ('@' in self.text and [*str(self.text)].count('@') >= 2):

            Tags: list = findall(r'\@(\w+)\@', self.text)

            # TODO: set object type
            return [{'type': 'MentionText', 'mention_text_object_guid': guid, 'from_index': from_index, 'length': len(length), 'mention_text_object_type': 'User'} for from_index, length, guid in zip(list(map(lambda i: self.text.index(i), Tags)), Tags, guids)], self.__replacer(self.text)


from html import escape # html escape 


def _add_surrogate(text):
    return ''.join(
        ''.join(chr(y) for y in struct.unpack('<HH', x.encode('utf-16le')))
        if (0x10000 <= ord(x) <= 0x10FFFF) else x for x in text
    )


def _del_surrogate(text):
    return text.encode('utf-16', 'surrogatepass').decode('utf-16')


class TypeMessageEntity(object):
    pass


class HTMLToTelegramParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.text = ''
        self.entities = []
        self._building_entities = {}
        self._open_tags = deque()
        self._open_tags_meta = deque()

    def handle_starttag(self, tag, attrs):
        self._open_tags.appendleft(tag)
        self._open_tags_meta.appendleft(None)

        attrs = dict(attrs)
        EntityType = None
        args = {}
        if tag == 'strong' or tag == 'b':
            EntityType = MessageEntityBold
        elif tag == 'em' or tag == 'i':
            EntityType = MessageEntityItalic
        elif tag == 'u':
            EntityType = MessageEntityUnderline
        elif tag == 'del' or tag == 's':
            EntityType = MessageEntityStrike
        elif tag == 'blockquote':
            EntityType = MessageEntityBlockquote
        elif tag == 'code':
            try:
                pre = self._building_entities['pre']
                try:
                    pre.language = attrs['class'][len('language-'):]
                except KeyError:
                    pass
            except KeyError:
                EntityType = MessageEntityCode
        elif tag == 'pre':
            EntityType = MessageEntityPre
            args['language'] = ''
        elif tag == 'a':
            try:
                url = attrs['href']
            except KeyError:
                return
            if url.startswith('mailto:'):
                url = url[len('mailto:'):]
                EntityType = MessageEntityEmail
            else:
                if self.get_starttag_text() == url:
                    EntityType = MessageEntityUrl
                else:
                    EntityType = MessageEntityTextUrl
                    args['url'] = _del_surrogate(url)
                    url = None
            self._open_tags_meta.popleft()
            self._open_tags_meta.appendleft(url)

        if EntityType and tag not in self._building_entities:
            self._building_entities[tag] = EntityType(
                offset=len(self.text),
                length=0,
                **args)

    def handle_data(self, text):
        previous_tag = self._open_tags[0] if len(self._open_tags) > 0 else ''
        if previous_tag == 'a':
            url = self._open_tags_meta[0]
            if url:
                text = url

        for tag, entity in self._building_entities.items():
            entity.length += len(text)

        self.text += text

    def handle_endtag(self, tag):
        try:
            self._open_tags.popleft()
            self._open_tags_meta.popleft()
        except IndexError:
            pass
        entity = self._building_entities.pop(tag, None)
        if entity:
            self.entities.append(entity)

def parse(html: str) -> typing.Tuple[str, typing.List[TypeMessageEntity]]:
    if not html:
        return html, []

    parser = HTMLToTelegramParser()
    parser.feed(_add_surrogate(html))
    text = helpers.strip_text(parser.text, parser.entities)
    return _del_surrogate(text), parser.entities

def unparse(text: str, entities: typing.Iterable[TypeMessageEntity], _offset: int = 0,
            _length: typing.Optional[int] = None) -> str:

    if not text:
        return text
    elif not entities:
        return escape(text)

    text = _add_surrogate(text)
    if _length is None:
        _length = len(text)
    html = []
    last_offset = 0
    for i, entity in enumerate(entities):
        if entity.offset >= _offset + _length:
            break
        relative_offset = entity.offset - _offset
        if relative_offset > last_offset:
            html.append(escape(text[last_offset:relative_offset]))
        elif relative_offset < last_offset:
            continue

        skip_entity = False
        length = entity.length

        while helpers.within_surrogate(text, relative_offset, length=_length):
            relative_offset += 1

        while helpers.within_surrogate(text, relative_offset + length, length=_length):
            length += 1

        entity_text = unparse(text=text[relative_offset:relative_offset + length],
                              entities=entities[i + 1:],
                              _offset=entity.offset, _length=length)
        entity_type = type(entity)

        if entity_type == MessageEntityBold:
            html.append('<strong>{}</strong>'.format(entity_text))
        elif entity_type == MessageEntityItalic:
            html.append('<em>{}</em>'.format(entity_text))
        elif entity_type == MessageEntityCode:
            html.append('<code>{}</code>'.format(entity_text))
        elif entity_type == MessageEntityUnderline:
            html.append('<u>{}</u>'.format(entity_text))
        elif entity_type == MessageEntityStrike:
            html.append('<del>{}</del>'.format(entity_text))
        elif entity_type == MessageEntityBlockquote:
            html.append('<blockquote>{}</blockquote>'.format(entity_text))
        elif entity_type == MessageEntityPre:
            if entity.language:
                html.append(
                    "<pre>\n"
                    "    <code class='language-{}'>\n"
                    "        {}\n"
                    "    </code>\n"
                    "</pre>".format(entity.language, entity_text))
            else:
                html.append('<pre><code>{}</code></pre>'
                            .format(entity_text))
        elif entity_type == MessageEntityEmail:
            html.append('<a href="mailto:{0}">{0}</a>'.format(entity_text))
        elif entity_type == MessageEntityUrl:
            html.append('<a href="{0}">{0}</a>'.format(entity_text))
        elif entity_type == MessageEntityTextUrl:
            html.append('<a href="{}">{}</a>'
                        .format(escape(entity.url), entity_text))
        elif entity_type == MessageEntityMentionName:
            html.append('<a href="tg://user?id={}">{}</a>'
                        .format(entity.user_id, entity_text))
        else:
            skip_entity = True
        last_offset = relative_offset + (0 if skip_entity else length)

    while helpers.within_surrogate(text, last_offset, length=_length):
        last_offset += 1

    html.append(escape(text[last_offset:]))
    return _del_surrogate(''.join(html))


class MetaDataLoader(object):
    
    @classmethod
    def __init__(cls, caption: str,
                 metadata: dict = None) -> (None):
        cls.text, cls.meta = caption, metadata
    
    def __max(cls) -> (dict):
        pass


#!/bin/python

import difflib, typing, logging, sys
from requests   import post, get, session as sessions
from .crypto    import Encryption

Client = 'RubikaClient'

class Rubino(object):

    def __init__(self, base_logger: str = __name__,
                 session: str = None) -> (None):
        
        '''
        from rb import RubinoClient

        with RubinoClient(...) as client:
            client.create_page(...)
        '''
        
        if isinstance(base_logger, str):
            base_logger: str = logging.getLogger(base_logger)
        
        elif not isinstance(base_logger, logging.Logger):
            
            try:
                from . import __name__ as base_log
                base_logger: str = base_log
            except Exception: ...

        self.session, self.auth, self.url, self.logger = sessions(), session, 'https://rubino12.iranlms.ir', self.base_logger

    def __call__(self):
        pass
    
    def __dir__(self):
        pass
    
    def __enter__(self) -> (object):
        return self

    def __exit__(
        self,
        *args,
        **kwargs
        ) -> (None):
        pass

    def __post(self, json: dict) -> (dict):
        
        for i in range(5):
            
            try:
                with self.session.post(
                    url=self.url,
                    json=json
                    ) as (res):
                    
                    if (res.status_code != 200):
                        continue

                    else:
                        return res.json()
                        break

            except Exception as e:
                print(e)

    def makeJson(
        self,
        method  :   str,
        data    :   dict
        ) -> (dict):

        return (
            {
                'api_version'   :   '0',
                'auth'          :   self.auth,
                'client'        :   clients.android,
                'data'          :   data,
                'method'        :   method
                }
        )

    def get_profile_list(
        self,
        limit   :   typing.Union[int, str]      =   10,
        sort    :   str                         =   'FromMax',
        equal   :   bool                        =   False
        ) -> (dict):
        
        json = self.makeJson(
            'getProfileList',
            {
                'equal': equal,
                'limit': limit,
                'sort' : sort,
                }
        )
        return self.__post(json=json)

    def request_follow(
        self,
        followee_id :   str,
        profile_id  :   str,
        f_type      :   str  =  'Follow'
        ) -> (dict):
        
        '''
        `f_type` is a action for follow type; and actions: `Follow` and `Unfollow`
        '''
        
        json = self.makeJson(
            'requestFollow',
            {
                'f_type'        :   'Follow',
                'followee_id'   :   followee_id,
                'profile_id'    :   profile_id
                }
            )

        return self.__post(json=json)

    def create_page(
        self, **kwargs
        ) -> (dict):
        
        '''
        `self.create_page(bio='', name='', username='', email='')`
        '''
        
        json = self.makeJson(
            'createPage',
            {**kwargs}
            )
        return self.__post(json=json)

    def update_profile(
        self,
        **kwargs
        ) -> (dict):
        
        '''
        `self.update_profile(bio='', name='', username='', email='')`
        '''
        json = self.makeJson(
            'updateProfile',
            {**kwargs}
            )
        
        return self.__post(json=json)

    def is_exist_username(
        self,
        username: str
        ) -> (dict):
        
        json = self.makeJson(
            'isExistUsername',
            {
                'username': username.replace('@','')
                }
            )
        return self.__post(json=json)

    def get_post_by_shareLink(
        self,
        share_string:   str,
        profile_id  :   str
        ) -> (dict):
        
        json = self.makeJson(
            'getPostByShareLink',
            {
                'share_string'  :   share_string,
                'profile_id'    :   profile_id
                }
            )
        return self.__post(json=json)

    def add_comment(
        self,
        text            :   str,
        post_id         :   str,
        post_profile_id :   str,
        profile_id      :   str
        ) -> (dict):
        
        json = self.makeJson(
            'addComment',
            {
                'content'           :   text,
                'post_id'           :   post_id,
                'post_profile_id'   :   post_profile_id,
                'rnd'               :   randint(1111111111, 9999999999),
                'profile_id'        :   profile_id
                }
            )

        return self.__post(json=json)

    def like_post_action(
        self,
        post_id         :   str,
        post_profile_id :   str,
        profile_id      :   str,
        action_type     :   str =   'Like' or 'Unlike'
        ) -> (dict):
        
        
        '''
        `action_type` is a action for post; and, actions: `Like` or `Unlike`
        '''

        json = self.makeJson(
            'likePostAction',
            {
                'action_type'       :   action_type,
                'post_id'           :   post_id,
                'post_profile_id'   :   post_profile_id,
                'profile_id'        :   profile_id
                }
            )

        return self.__post(json=json)

    def add_post_view_count(
        self,
        post_id         :   str,
        post_profile_id :   str
        ) -> (dict):
        
        json = self.makeJson(
            'addPostViewCount',
            {
                'post_id'           :   post_id,
                'post_profile_id'   :   post_profile_id
                }
            )
        return self.__post(json=json)

    def get_comments(self,
        post_id         :   str,
        profile_id      :   str,
        post_profile_id :   str,
        limit           :   int     =   50,
        sort            :   str     =   'FromMax',
        equal           :   bool    =   False
        ) -> (dict):
        
        json = self.makeJson(
            'getComments',
            {
                'equal'             :   equal,
                'limit'             :   limit,
                'sort'              :   sort,
                'post_id'           :   post_id,
                'profile_id'        :   profile_id,
                'post_profile_id'   :   post_profile_id
                }
            )
        return self.__post(json=json)

    def get_recent_following_posts(
        self        :   'RubinoClient',
        profile_id  :   'str',
        limit       :   'int'   =   30,
        sort        :   'str'   =   'FromMax',
        equal       :   'bool'  =   False
        ) -> (dict):
        
        json = self.makeJson(
            'getRecentFollowingPosts',
            {
            'equal'         :   equal,
            'limit'         :   limit,
            'sort'          :   sort,
            'profile_id'    :   profile_id
            }
            )
        return self.__post(json=json)

    def get_profile_posts(self,
        target_profile_id   :   str,
        profile_id          :   str,
        limit               :   int     =   50,
        sort                :   str     =   'FromMax',
        equal               :   bool    =   False
        ) -> (dict):
        
        json = self.makeJson(
            'getRecentFollowingPosts',
            {
                'equal'             :   equal,
                'limit'             :   limit,
                'sort'              :   sort,
                'profile_id'        :   profile_id,
                'target_profile_id' :   target_profile_id
                }
            )
        return self.__post(json=json)

    def get_profile_stories(
        self,
        target_profile_id   :   str,
        limit               :   int =   100
        ) -> (dict):
        
        json = self.makeJson(
            'getProfileStories',
            {
                'limit'         :   limit,
                'profile_id'    :   target_profile_id
                }
            )
        return self.__post(json=json)

class RubinoClient(Rubino):
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        
        try:
            sys.modules[self.base_logger] = RubinoClient
        except Exception: ...

    def __getattr__(self, name, *args, **kwargs) -> typing.Union[object, None]:
        
        method = 'MethodNameNotFound'
        
        attr = difflib.get_close_matches(name, dir(Rubino), n=1)

        if attr:
            method = getattr(__base[0], attr[0])
        
        try:
            return method(*args, **kwargs)
        except Exception:
            return method